# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from typing import TYPE_CHECKING

from .equine import Equine, EquineOutput
from .equine_gp import EquineGP
from .equine_protonet import CovType, EquineProtonet
from .load_equine_model import load_equine_model
from .utils import (
    brier_score,
    brier_skill_score,
    expected_calibration_error,
    generate_episode,
    generate_model_metrics,
    generate_model_summary,
    generate_support,
    generate_train_summary,
    mahalanobis_distance_nosq,
)

if not TYPE_CHECKING:  # pragma: no cover
    try:
        from ._version import version as __version__
    except ImportError:
        __version__ = "unknown version"
else:  # pragma: no cover
    __version__: str

__all__ = [
    "Equine",
    "EquineOutput",
    "EquineGP",
    "EquineProtonet",
    "CovType",
    "brier_score",
    "brier_skill_score",
    "expected_calibration_error",
    "load_equine_model",
    "generate_support",
    "generate_episode",
    "generate_model_metrics",
    "generate_train_summary",
    "generate_model_summary",
    "mahalanobis_distance_nosq",
]



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

import os
import torch
from conftest import (
    generate_random_string_list,
    random_dataset,
    use_basic_embedding_model,
    use_save_load_model_tests,
)
from hypothesis import given, settings

import equine as eq


@given(random_dataset=random_dataset())
@settings(deadline=None, max_examples=10)
def test_equine_gp_train_from_scratch(random_dataset) -> None:
    dataset, num_classes, X, embedding_model = use_basic_embedding_model(random_dataset)

    model = eq.EquineGP(embedding_model, num_classes, num_classes)
    loss_fn = torch.nn.CrossEntropyLoss()
    optimizer = torch.optim.SGD(
        model.parameters(),
        lr=0.001,
        momentum=0.9,
        weight_decay=0.0001,
    )
    model.train_model(dataset, loss_fn, optimizer, num_epochs=10)

    model.predict(X[1:10])  # Contracts should fire asserts on errors


@given(random_dataset=random_dataset())
@settings(deadline=None, max_examples=10)
def test_equine_gp_train_from_scratch_with_temperature(random_dataset) -> None:
    dataset, num_classes, X, embedding_model = use_basic_embedding_model(random_dataset)

    model = eq.EquineGP(embedding_model, num_classes, num_classes, use_temperature=True)
    loss_fn = torch.nn.CrossEntropyLoss()
    optimizer = torch.optim.SGD(
        model.parameters(),
        lr=0.001,
        momentum=0.9,
        weight_decay=0.0001,
    )
    train_dict = model.train_model(dataset, loss_fn, optimizer, num_epochs=10)
    assert "calibration_loader" in train_dict
    assert train_dict["calibration_loader"] is not None
    cal_loader = train_dict["calibration_loader"]

    model.calibrate_temperature(cal_loader, 1, 0.01)

    model.predict(X[1:10])  # Contracts should fire asserts on errors


@given(random_dataset=random_dataset())
@settings(deadline=None, max_examples=2)
def test_equine_gp_save_load(random_dataset) -> None:
    dataset, num_classes, X, embedding_model = use_basic_embedding_model(random_dataset)

    model = eq.EquineGP(embedding_model, num_classes, num_classes)
    loss_fn = torch.nn.CrossEntropyLoss()
    optimizer = torch.optim.SGD(
        model.parameters(),
        lr=0.001,
        momentum=0.9,
        weight_decay=0.0001,
    )
    model.train_model(dataset, loss_fn, optimizer, num_epochs=2)

    new_model, tmp_filename = use_save_load_model_tests(
        model, X, tmp_filename="gp_save_load.eq"
    )

    if os.path.exists(tmp_filename):
        os.remove(tmp_filename)  # Cleanup


@given(random_dataset=random_dataset())
@settings(deadline=None, max_examples=1)
def test_equine_gp_save_load_with_temperature(random_dataset) -> None:
    dataset, num_classes, X, embedding_model = use_basic_embedding_model(random_dataset)

    model = eq.EquineGP(embedding_model, num_classes, num_classes, use_temperature=True)
    loss_fn = torch.nn.CrossEntropyLoss()
    optimizer = torch.optim.SGD(
        model.parameters(),
        lr=0.001,
        momentum=0.9,
        weight_decay=0.0001,
    )
    model.train_model(dataset, loss_fn, optimizer, num_epochs=2)

    new_model, tmp_filename = use_save_load_model_tests(
        model, X, tmp_filename="gp_save_load_with_temperature.eq"
    )

    if os.path.exists(tmp_filename):
        os.remove(tmp_filename)  # Cleanup


@given(random_dataset=random_dataset())
@settings(deadline=None, max_examples=1)
def test_equine_gp_save_load_with_vis(random_dataset) -> None:
    dataset, num_classes, X, embedding_model = use_basic_embedding_model(random_dataset)

    model = eq.EquineGP(embedding_model, num_classes, num_classes)
    loss_fn = torch.nn.CrossEntropyLoss()
    optimizer = torch.optim.SGD(
        model.parameters(),
        lr=0.001,
        momentum=0.9,
        weight_decay=0.0001,
    )
    model.train_model(dataset, loss_fn, optimizer, num_epochs=2, vis_support=True)

    new_model, tmp_filename = use_save_load_model_tests(
        model, X, tmp_filename="gp_save_load_with_vis.eq"
    )

    assert new_model.support is not None, "support was not saved"
    assert new_model.prototypes is not None, "prototypes were not saved"
    assert (
        model.support.keys() == new_model.get_support().keys()
    ), "Support keys changed on reload"
    assert (
        torch.nn.functional.mse_loss(model.prototypes, new_model.get_prototypes())
        <= 1e-7
    ), "Prototypes changed on reload"

    if os.path.exists(tmp_filename):
        os.remove(tmp_filename)  # Cleanup


@given(random_dataset=random_dataset())
@settings(deadline=None, max_examples=1)
def test_equine_gp_save_load_with_feature_and_label_names(random_dataset) -> None:
    dataset, num_classes, X, embedding_model = use_basic_embedding_model(random_dataset)

    # without feature and label names
    model = eq.EquineGP(embedding_model, num_classes, num_classes)
    loss_fn = torch.nn.CrossEntropyLoss()
    optimizer = torch.optim.SGD(
        model.parameters(),
        lr=0.001,
        momentum=0.9,
        weight_decay=0.0001,
    )
    model.train_model(dataset, loss_fn, optimizer, num_epochs=2)

    new_model, tmp_filename = use_save_load_model_tests(
        model, X, tmp_filename="gp_save_load_no_feature_and_label_names.eq"
    )

    assert new_model.get_feature_names() is None, "feature_names changed on reload"
    assert new_model.get_label_names() is None, "label_names changed on reload"

    if os.path.exists(tmp_filename):
        os.remove(tmp_filename)  # Cleanup

    # with feature and label names
    feature_names = generate_random_string_list(X.shape[1])
    label_names = generate_random_string_list(num_classes)

    model = eq.EquineGP(
        embedding_model,
        num_classes,
        num_classes,
        feature_names=feature_names,
        label_names=label_names,
    )
    loss_fn = torch.nn.CrossEntropyLoss()
    optimizer = torch.optim.SGD(
        model.parameters(),
        lr=0.001,
        momentum=0.9,
        weight_decay=0.0001,
    )
    model.train_model(dataset, loss_fn, optimizer, num_epochs=2)

    new_model, tmp_filename = use_save_load_model_tests(
        model, X, tmp_filename="gp_save_load_with_feature_and_label_names.eq"
    )

    assert (
        new_model.get_feature_names() == feature_names
    ), "feature_names changed on reload"
    assert new_model.get_label_names() == label_names, "label_names changed on reload"

    if os.path.exists(tmp_filename):
        os.remove(tmp_filename)  # Cleanup



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT
from __future__ import annotations

from typing import Any, Optional

import icontract
import io
import numpy as np
import torch
import warnings
from beartype import beartype
from collections import OrderedDict
from collections.abc import Callable
from datetime import datetime
from enum import Enum
from scipy.stats import gaussian_kde
from torch.utils.data import TensorDataset
from tqdm import tqdm

from .equine import Equine, EquineOutput
from .utils import (
    generate_episode,
    generate_support,
    generate_train_summary,
    mahalanobis_distance_nosq,
    stratified_train_test_split,
)


#####################################
class CovType(Enum):
    """
    Enum class for covariance types used in EQUINE.
    """

    UNIT = "unit"
    DIAGONAL = "diag"
    FULL = "full"


PRED_COV_TYPE = CovType.DIAGONAL
OOD_COV_TYPE = CovType.DIAGONAL
DEFAULT_EPSILON = 1e-5
COV_REG_TYPE = "epsilon"


###############################################


@beartype
class Protonet(torch.nn.Module):
    """
    Private class that implements a prototypical neural network for use in EQUINE.
    """

    def __init__(
        self,
        embedding_model: torch.nn.Module,
        emb_out_dim: int,
        cov_type: CovType,
        cov_reg_type: str,
        epsilon: float,
        device: str = "cpu",
    ) -> None:
        """
        Protonet class constructor.

        Parameters
        ----------
        embedding_model : torch.nn.Module
            The PyTorch embedding model to generate logits with.
        emb_out_dim : int
            Dimension size of given embedding model's output.
        cov_type : CovType
            Type of covariance to use when computing distances [unit, diag, full].
        cov_reg_type : str
            Type of regularization to use when generating the covariance matrix [epsilon, shared].
        epsilon : float
            Epsilon value to use for covariance regularization.
        device : str, optional
            The device to train the protonet model on (defaults to cpu).
        """
        super().__init__()
        self.embedding_model = embedding_model
        self.cov_type = cov_type
        self.cov_reg_type = cov_reg_type
        self.epsilon = epsilon
        self.emb_out_dim = emb_out_dim
        self.to(device)
        self.device = device

        self.support: OrderedDict[int, torch.Tensor] = OrderedDict()
        self.support_embeddings: OrderedDict[int, torch.Tensor] = OrderedDict()
        self.model_head: torch.nn.Module = self.create_model_head(emb_out_dim)
        self.model_head.to(device)

    def create_model_head(self, emb_out_dim: int) -> torch.nn.Linear:
        """
        Method for adding a PyTorch layer on top of the given embedding model. This layer
        is intended to offer extra degrees of freedom for distance learning in the embedding space.

        Parameters
        ----------
        emb_out_dim : int
            Dimension size of the embedding model output.

        Returns
        -------
        torch.nn.Linear
            The created PyTorch model layer.
        """
        return torch.nn.Linear(emb_out_dim, emb_out_dim)

    def compute_embeddings(self, X: torch.Tensor) -> torch.Tensor:
        """
        Method for calculating model embeddings using both the given embedding model and the added model head.

        Parameters
        ----------
        X : torch.Tensor
            Input tensor to compute embeddings on.

        Returns
        -------
        torch.Tensor
            Fully computed embedding tensors for the given X tensor.
        """
        model_embeddings = self.embedding_model(X.to(self.device))
        head_embeddings = self.model_head(model_embeddings)
        return head_embeddings

    @icontract.require(lambda self: len(self.support_embeddings) > 0)
    def compute_prototypes(self) -> torch.Tensor:
        """
        Method for computing class prototypes based on given support examples.
        ``Prototypes'' in this context are the means of the support embeddings for each class.

        Returns
        -------
        torch.Tensor
            Tensors of prototypes for each of the given classes in the support.
        """
        # Compute prototype for each class
        proto_list = []
        for label in self.support_embeddings:  # look at doing functorch
            class_prototype = torch.mean(self.support_embeddings[label], dim=0)
            proto_list.append(class_prototype)

        prototypes = torch.stack(proto_list)

        return prototypes

    @icontract.require(lambda self: len(self.support_embeddings) > 0)
    def compute_covariance(self, cov_type: CovType) -> torch.Tensor:
        """
        Method for generating the (regularized) support example covariance matrix(es) used for calculating distances.
        Note that this method is only called once per episode, and the resulting tensor is used for all queries.

        Parameters
        ----------
        cov_type : CovType
            Type of covariance to use [unit, diag, full].

        Returns
        -------
        torch.Tensor
            Tensor containing the generated regularized covariance matrix.
        """
        class_cov_dict = OrderedDict().fromkeys(
            self.support_embeddings.keys(), torch.Tensor()
        )
        for label in self.support_embeddings.keys():
            class_covariance = self.compute_covariance_by_type(
                cov_type, self.support_embeddings[label]
            )
            class_cov_dict[label] = class_covariance

        reg_covariance_dict = self.regularize_covariance(
            class_cov_dict, cov_type, self.cov_reg_type
        )
        reg_covariance = torch.stack(list(reg_covariance_dict.values()))

        return reg_covariance  # TODO try putting everything on GPU with .to() and see if faster

    def compute_covariance_by_type(
        self, cov_type: CovType, embedding: torch.Tensor
    ) -> torch.Tensor:
        """
        Select the appropriate covariance matrix type based on cov_type.

        Parameters
        ----------
        cov_type : str
            Type of covariance to use. Options are ['unit', 'diag', 'full'].
        embedding : torch.Tensor
            Embedding tensor to use when generating the covariance matrix.

        Returns
        -------
        torch.Tensor
            Tensor containing the requested covariance matrix.
        """
        if cov_type == CovType.FULL:
            class_covariance = torch.cov(embedding.T)
        elif cov_type == CovType.DIAGONAL:
            class_covariance = torch.var(embedding, dim=0)
        elif cov_type == CovType.UNIT:
            class_covariance = torch.ones(self.emb_out_dim)
        else:
            raise ValueError

        return class_covariance

    def regularize_covariance(
        self,
        class_cov_dict: OrderedDict[int, torch.Tensor],
        cov_type: CovType,
        cov_reg_type: str,
    ) -> OrderedDict[int, torch.Tensor]:
        """
        Method to add regularization to each class covariance matrix based on the selected regularization type.

        Parameters
        ----------
        class_cov_dict : OrderedDict[int, torch.Tensor]
            A dictionary containing each class and the corresponding covariance matrix.
        cov_type : CovType
            Type of covariance to use [unit, diag, full].

        Returns
        -------
        dict[float, torch.Tensor]
            Dictionary containing the regularized class covariance matrices.
        """

        if cov_type == CovType.FULL:
            regularization = torch.diag(self.epsilon * torch.ones(self.emb_out_dim)).to(
                self.device
            )
        elif cov_type == CovType.DIAGONAL:
            regularization = self.epsilon * torch.ones(self.emb_out_dim).to(self.device)
        elif cov_type == CovType.UNIT:
            regularization = torch.zeros(self.emb_out_dim).to(self.device)

        if cov_reg_type == "shared":
            if cov_type != CovType.FULL and cov_type != CovType.DIAGONAL:
                for label in self.support_embeddings:
                    class_cov_dict[label] = class_cov_dict[label] + regularization
                warnings.warn(
                    "Covariance type UNIT is incompatible with shared regularization, \
                    reverting to epsilon regularization"
                )
                return class_cov_dict

            shared_covariance = self.compute_shared_covariance(class_cov_dict, cov_type)

            for label in self.support_embeddings:
                num_class_support = self.support_embeddings[label].shape[0]
                lamb = num_class_support / (num_class_support + 1)

                class_cov_dict[label] = (
                    lamb * class_cov_dict[label]
                    + (1 - lamb) * shared_covariance
                    + regularization
                )

        elif cov_reg_type == "epsilon":
            for label in class_cov_dict.keys():
                class_cov_dict[label] = (
                    class_cov_dict[label].to(self.device) + regularization
                )

        return class_cov_dict

    def compute_shared_covariance(
        self, class_cov_dict: OrderedDict[int, torch.Tensor], cov_type: CovType
    ) -> torch.Tensor:
        """
        Method to calculate a shared covariance matrix.

        The shared covariance matrix is calculated as the weighted average of the class covariance matrices,
        where the weights are the number of support examples for each class. This is useful when the number of
        support examples for each class is small.

        Parameters
        ----------
        class_cov_dict : OrderedDict[int, torch.Tensor]
            A dictionary containing each class and the corresponding covariance matrix.
        cov_type : CovType
            Type of covariance to use [unit, diag, full].

        Returns
        -------
        torch.Tensor
            Tensor containing the shared covariance matrix.
        """
        total_support = sum([x.shape[0] for x in class_cov_dict.values()])

        if cov_type == CovType.FULL:
            shared_covariance = torch.zeros((self.emb_out_dim, self.emb_out_dim))
        elif cov_type == CovType.DIAGONAL:
            shared_covariance = torch.zeros(self.emb_out_dim)
        else:
            raise ValueError(
                "Shared covariance can only be used with FULL or DIAGONAL (not UNIT) covariance types"
            )

        for label in class_cov_dict:
            num_class_support = class_cov_dict[label].shape[0]
            shared_covariance = (
                shared_covariance + (num_class_support - 1) * class_cov_dict[label]
            )  # undo N-1 div from cov

        shared_covariance = shared_covariance / (
            total_support - 1
        )  # redo N-1 div for shared cov

        return shared_covariance

    @icontract.require(lambda X_embed, mu: X_embed.shape[-1] == mu.shape[-1])
    @icontract.ensure(lambda result: torch.all(result >= 0))
    def compute_distance(
        self, X_embed: torch.Tensor, mu: torch.Tensor, cov: torch.Tensor
    ) -> torch.Tensor:
        """
        Method to compute the distances to class prototypes for the given embeddings.

        Parameters
        ----------
        X_embed : torch.Tensor
            The embeddings of the query examples.
        mu : torch.Tensor
            The class prototypes (means of the support embeddings).
        cov : torch.Tensor
            The support covariance matrix.

        Returns
        -------
        torch.Tensor
            The calculated distances from each of the class prototypes for the given embeddings.
        """
        _queries = torch.unsqueeze(X_embed, 1)  # examples x 1 x dimension
        diff = torch.sub(mu, _queries)

        if len(cov.shape) == 2:  # (diagonal covariance)
            # examples x classes x dimension
            sq_diff = diff**2
            div = torch.div(sq_diff.to(self.device), cov.to(self.device))
            dist = torch.nan_to_num(div)
            dist = torch.sum(dist, dim=2)  # examples x classes
            dist = dist.squeeze(dim=1)
            dist = torch.sqrt(dist + self.epsilon)  # examples x classes
        else:  # len(cov.shape) == 3: (full covariance)
            diff = diff.permute(1, 2, 0)  # classes x dimension x examples
            dist = mahalanobis_distance_nosq(diff, cov)
            dist = torch.sqrt(dist.permute(1, 0) + self.epsilon)  # examples x classes
            dist = dist.squeeze(dim=1)
        return dist

    def compute_classes(self, distances: torch.Tensor) -> torch.Tensor:
        """
        Method to compute predicted classes from distances via a softmax function.

        Parameters
        ----------
        distances : torch.Tensor
            The distances of embeddings to class prototypes.

        Returns
        -------
        torch.Tensor
            Tensor of class predictions.
        """
        softmax = torch.nn.functional.softmax(torch.neg(distances), dim=-1)
        return softmax

    def forward(self, X: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
        """
        Protonet forward function, generates class probability predictions and distances from prototypes.

        Parameters
        ----------
        X : torch.Tensor
            Input tensor of queries for generating predictions.

        Returns
        -------
        tuple[torch.Tensor, torch.Tensor]
            tuple containing class probability predictions, and class distances from prototypes.
        """
        if len(self.support) == 0 or len(self.support_embeddings) == 0:
            raise ValueError(
                "No support examples found. Protonet Model requires model support to \
                    be set with the 'update_support()' method before calling forward."
            )

        X_embed = self.compute_embeddings(X)
        if X_embed.shape == torch.Size([self.emb_out_dim]):
            X_embed = X_embed.unsqueeze(dim=0)  # handle single examples
        distances = self.compute_distance(X_embed, self.prototypes, self.covariance)
        classes = self.compute_classes(distances)

        return classes, distances

    def update_support(self, support: OrderedDict[int, torch.Tensor]) -> None:
        """
        Method to update the support examples, and all the calculations that rely on them.

        Parameters
        ----------
        support : OrderedDict
            Ordered dict containing class labels and their associated support examples.
        """
        self.support = support  # TODO torch.nn.ParameterDict(support)

        support_embs = OrderedDict().fromkeys(support.keys(), torch.Tensor())
        for label in support:
            support_embs[label] = self.compute_embeddings(support[label])

        self.support_embeddings = (
            support_embs  # TODO torch.nn.ParameterDict(support_embs)
        )

        self.prototypes: torch.Tensor = self.compute_prototypes()

        if self.training is False:
            self.compute_global_moments()
            self.covariance: torch.Tensor = self.compute_covariance(
                cov_type=PRED_COV_TYPE
            )
        else:
            self.covariance: torch.Tensor = self.compute_covariance(
                cov_type=self.cov_type
            )

    @icontract.require(lambda self: len(self.support_embeddings) > 0)
    def compute_global_moments(self) -> None:
        """Method to calculate the global moments of the support embeddings for use in OOD score generation"""
        embeddings = torch.cat(list(self.support_embeddings.values()))
        self.global_covariance = torch.unsqueeze(
            self.compute_covariance_by_type(OOD_COV_TYPE, embeddings), dim=0
        )
        global_reg_input = OrderedDict().fromkeys([0], torch.Tensor())
        global_reg_input[0] = self.global_covariance
        self.global_covariance: torch.Tensor = self.regularize_covariance(
            global_reg_input, OOD_COV_TYPE, "epsilon"
        )[0]
        self.global_mean: torch.Tensor = torch.mean(embeddings, dim=0)


###############################################
@beartype
class EquineProtonet(Equine):
    """
    A class representing an EQUINE model that utilizes protonets and (optionally) relative Mahalanobis distances
    to generate OOD and model confidence scores. This wraps any pytorch embedding neural network
    and provides the `forward`, `predict`, `save`, and `load` methods required by Equine.
    """

    def __init__(
        self,
        embedding_model: torch.nn.Module,
        emb_out_dim: int,
        cov_type: CovType = CovType.UNIT,
        relative_mahal: bool = True,
        use_temperature: bool = False,
        init_temperature: float = 1.0,
        device: str = "cpu",
        feature_names: Optional[list[str]] = None,
        label_names: Optional[list[str]] = None,
    ) -> None:
        """
        EquineProtonet class constructor

        Parameters
        ----------
        embedding_model : torch.nn.Module
            Neural Network feature embedding model.
        emb_out_dim : int
            The number of output features from the embedding model.
        cov_type : CovType, optional
            The type of covariance to use when training the protonet [UNIT, DIAG, FULL], by default CovType.UNIT.
        relative_mahal : bool, optional
            Use relative mahalanobis distance for OOD calculations. If false, uses standard mahalanobis distance instead, by default True.
        use_temperature : bool, optional
            Whether to use temperature scaling after training, by default False.
        init_temperature : float, optional
            What to use as the initial temperature (1.0 has no effect), by default 1.0.
        device : str, optional
            The device to train the equine model on (defaults to cpu).
        feature_names : list[str], optional
            List of strings of the names of the tabular features (ex ["duration", "fiat_mean", ...])
        label_names : list[str], optional
            List of strings of the names of the labels (ex ["streaming", "voip", ...])
        """
        super().__init__(
            embedding_model,
            device=device,
            feature_names=feature_names,
            label_names=label_names,
        )
        self.cov_type = cov_type
        self.cov_reg_type = COV_REG_TYPE
        self.relative_mahal = relative_mahal
        self.emb_out_dim = emb_out_dim
        self.epsilon = DEFAULT_EPSILON
        self.outlier_score_kde: OrderedDict[int, gaussian_kde] = OrderedDict()
        self.model_summary: dict[str, Any] = dict()
        self.use_temperature = use_temperature
        self.init_temperature = init_temperature
        self.register_buffer(
            "temperature", torch.Tensor(self.init_temperature * torch.ones(1))
        )

        self.model: torch.nn.Module = Protonet(
            embedding_model,
            self.emb_out_dim,
            self.cov_type,
            self.cov_reg_type,
            self.epsilon,
            device=device,
        )

    def forward(self, X: torch.Tensor) -> torch.Tensor:
        """
        Generates logits for classification based on the input tensor.

        Parameters
        ----------
        X : torch.Tensor
            The input tensor for generating predictions.

        Returns
        -------
        torch.Tensor
            The output class predictions.
        """
        preds, _ = self.model(X)
        return preds

    @icontract.require(lambda calib_frac: calib_frac > 0 and calib_frac < 1)
    def train_model(
        self,
        dataset: TensorDataset,
        num_episodes: int,
        calib_frac: float = 0.2,
        support_size: int = 25,
        way: int = 3,
        episode_size: int = 100,
        loss_fn: Callable = torch.nn.functional.cross_entropy,
        opt_class: Callable = torch.optim.Adam,
        num_calibration_epochs: int = 2,
        calibration_lr: float = 0.01,
    ) -> dict[str, Any]:
        """
        Train or fine-tune an EquineProtonet model.

        Parameters
        ----------
        dataset : TensorDataset
            Input pytorch TensorDataset of training data for model.
        num_episodes : int
            The desired number of episodes to use for training.
        calib_frac : float, optional
            Fraction of given training data to reserve for model calibration, by default 0.2.
        support_size : int, optional
            Number of support examples to generate for each class, by default 25.
        way : int, optional
            Number of classes to train on per episode, by default 3.
        episode_size : int, optional
            Number of examples to use per episode, by default 100.
        loss_fn : Callable, optional
            A pytorch loss function, eg., torch.nn.CrossEntropyLoss(), by default torch.nn.functional.cross_entropy.
        opt_class : Callable, optional
            A pytorch optimizer, e.g., torch.optim.Adam, by default torch.optim.Adam.
        num_calibration_epochs : int, optional
            The desired number of epochs to use for temperature scaling, by default 2.
        calibration_lr : float, optional
            Learning rate for temperature scaling, by default 0.01.

        Returns
        -------
        tuple[dict[str, Any], torch.Tensor, torch.Tensor]
            A tuple containing the model summary, the held out calibration data, and the calibration labels.
        """
        self.train()

        if self.use_temperature:
            self.temperature: torch.Tensor = torch.Tensor(
                self.init_temperature * torch.ones(1)
            ).type_as(self.temperature)

        X, Y = dataset[:]

        self.validate_feature_label_names(X.shape[-1], torch.unique(Y).shape[0])

        train_x, calib_x, train_y, calib_y = stratified_train_test_split(
            X, Y, test_size=calib_frac
        )
        optimizer = opt_class(self.parameters())

        train_x.to(self.device)
        train_y.to(self.device)
        calib_x.to(self.device)
        calib_y.to(self.device)

        for i in tqdm(range(num_episodes)):
            optimizer.zero_grad()

            support, episode_x, episode_y = generate_episode(
                train_x, train_y, support_size, way, episode_size
            )
            self.model.update_support(support)

            _, dists = self.model(episode_x)
            loss_value = loss_fn(
                torch.neg(dists).to(self.device), episode_y.to(self.device)
            )
            loss_value.backward()
            optimizer.step()

        self.eval()
        full_support = generate_support(
            train_x,
            train_y,
            support_size,
            selected_labels=torch.unique(train_y).tolist(),
        )

        self.model.update_support(
            full_support
        )  # update support with final selected examples

        X_embed = self.model.compute_embeddings(calib_x)
        pred_probs, dists = self.model(calib_x)
        ood_dists = self._compute_ood_dist(X_embed, pred_probs, dists)
        self._fit_outlier_scores(ood_dists, calib_y)

        if self.use_temperature:
            self.calibrate_temperature(
                calib_x, calib_y, num_calibration_epochs, calibration_lr
            )

        date_trained = datetime.now().strftime("%m/%d/%Y, %H:%M:%S")
        self.train_summary: dict[str, Any] = generate_train_summary(
            self, train_y, date_trained
        )
        return_dict: dict[str, Any] = dict()
        return_dict["train_summary"] = self.train_summary
        return_dict["calib_x"] = calib_x
        return_dict["calib_y"] = calib_y
        return return_dict

    def calibrate_temperature(
        self,
        calib_x: torch.Tensor,
        calib_y: torch.Tensor,
        num_calibration_epochs: int = 1,
        calibration_lr: float = 0.01,
    ) -> None:
        """
        Fine-tune the temperature after training. Note that this function is also run at the conclusion of train_model.

        Parameters
        ----------
        calib_x : torch.Tensor
            Training data to be used for temperature calibration.
        calib_y : torch.Tensor
            Labels corresponding to `calib_x`.
        num_calibration_epochs : int, optional
            Number of epochs to tune temperature, by default 1.
        calibration_lr : float, optional
            Learning rate for temperature optimization, by default 0.01.

        Returns
        -------
        None
        """
        self.temperature.requires_grad = True
        optimizer = torch.optim.Adam([self.temperature], lr=calibration_lr)
        for t in range(num_calibration_epochs):
            optimizer.zero_grad()
            with torch.no_grad():
                pred_probs, dists = self.model(calib_x)
            dists = dists.to(self.device) / self.temperature.to(self.device)
            loss = torch.nn.functional.cross_entropy(
                torch.neg(dists).to(self.device), calib_y.to(torch.long).to(self.device)
            )
            loss.backward()
            optimizer.step()
        self.temperature.requires_grad = False

    @icontract.ensure(lambda self: len(self.model.support_embeddings) > 0)
    def _fit_outlier_scores(
        self, ood_dists: torch.Tensor, calib_y: torch.Tensor
    ) -> None:
        """
        Private function to fit outlier scores with a kernel density estimate (KDE).

        Parameters
        ----------
        ood_dists : torch.Tensor
            Tensor of computed OOD distances.
        calib_y : torch.Tensor
            Tensor of class labels for `ood_dists` examples.

        Returns
        -------
        None
        """
        for label in self.model.support_embeddings.keys():
            class_ood_dists = ood_dists[calib_y == int(label)].cpu().detach().numpy()
            class_kde = gaussian_kde(class_ood_dists)  # TODO convert to torch func
            self.outlier_score_kde[label] = class_kde

    def _compute_outlier_scores(self, ood_dists, predictions) -> torch.Tensor:
        """
        Private function to compute OOD scores using the calculated kernel density estimate (KDE).

        Parameters
        ----------
        ood_dists : torch.Tensor
            Tensor of computed OOD distances.
        predictions : torch.Tensor
            Tensor of model protonet predictions.

        Returns
        -------
        torch.Tensor
            Tensor of OOD scores for the given examples.
        """
        ood_scores = torch.zeros_like(ood_dists)
        for i in range(len(predictions)):
            # Use KDE and RMD corresponding to the predicted class
            predicted_class = int(torch.argmax(predictions[i, :]))
            p_value = self.outlier_score_kde[int(predicted_class)].integrate_box_1d(
                ood_dists[i].detach().numpy(), np.inf
            )
            ood_scores[i] = 1.0 - np.clip(p_value, 0.0, 1.0)

        return ood_scores

    @icontract.ensure(lambda result: len(result) > 0)
    def _compute_ood_dist(
        self,
        X_embeddings: torch.Tensor,
        predictions: torch.Tensor,
        distances: torch.Tensor,
    ) -> torch.Tensor:
        """
        Private function to compute OOD distances using a distance function.

        Parameters
        ----------
        X_embeddings : torch.Tensor
            Tensor of example embeddings.
        predictions : torch.Tensor
            Tensor of model protonet predictions for the given embeddings.
        distances : torch.Tensor
            Tensor of calculated protonet distances for the given embeddings.

        Returns
        -------
        torch.Tensor
            Tensor of OOD distances for the given embeddings.
        """
        preds = torch.argmax(predictions, dim=1)
        preds = preds.unsqueeze(dim=-1)
        # Calculate (Relative) Mahalanobis Distance:
        if self.relative_mahal:
            null_distance = self.model.compute_distance(
                X_embeddings, self.model.global_mean, self.model.global_covariance
            )
            null_distance = null_distance.unsqueeze(dim=-1)
            ood_dist = distances.gather(1, preds) - null_distance
        else:
            ood_dist = distances.gather(1, preds)

        ood_dist = torch.reshape(ood_dist, (-1,))
        return ood_dist

    def predict(self, X: torch.Tensor) -> EquineOutput:
        """Predict function for EquineProtonet, inherited and implemented from Equine.

        Parameters
        ----------
        X : torch.Tensor
            Input tensor.

        Returns
        -------
        EquineOutput
            Output object containing prediction probabilities and OOD scores.
        """
        X_embed = self.model.compute_embeddings(X)
        if X_embed.shape == torch.Size([self.model.emb_out_dim]):
            X_embed = X_embed.unsqueeze(dim=0)  # Handle single examples
        preds, dists = self.model(X)
        if self.use_temperature:
            dists = dists / self.temperature
            preds = torch.softmax(torch.negative(dists), dim=1)
        ood_dist = self._compute_ood_dist(X_embed, preds, dists)
        ood_scores = self._compute_outlier_scores(ood_dist, preds)

        self.validate_feature_label_names(X.shape[-1], preds.shape[-1])

        return EquineOutput(classes=preds, ood_scores=ood_scores, embeddings=X_embed)

    @icontract.require(lambda calib_frac: (calib_frac > 0.0) and (calib_frac < 1.0))
    def update_support(
        self,
        support_x: torch.Tensor,
        support_y: torch.Tensor,
        calib_frac: float,
        label_names: Optional[list[str]] = None,
    ) -> None:
        """Function to update protonet support examples with given examples.

        Parameters
        ----------
        support_x : torch.Tensor
            Tensor containing support examples for protonet.
        support_y : torch.Tensor
            Tensor containing labels for given support examples.
        calib_frac : float
            Fraction of given support data to use for OOD calibration.
        label_names : list[str], optional
            List of strings of the names of the labels (ex ["streaming", "voip", ...])

        Returns
        -------
        None
        """

        support_x, calib_x, support_y, calib_y = stratified_train_test_split(
            support_x, support_y, test_size=calib_frac
        )
        labels, counts = torch.unique(support_y, return_counts=True)
        if label_names is not None:
            self.label_names = label_names
        self.validate_feature_label_names(support_x.shape[-1], labels.shape[0])

        support = OrderedDict()
        for label, count in list(zip(labels.tolist(), counts.tolist())):
            class_support = generate_support(
                support_x,
                support_y,
                support_size=count,
                selected_labels=[label],
            )
            support.update(class_support)

        self.model.update_support(support)

        X_embed = self.model.compute_embeddings(calib_x)
        preds, dists = self.model(calib_x)
        ood_dists = self._compute_ood_dist(X_embed, preds, dists)

        self._fit_outlier_scores(ood_dists, calib_y)

    @icontract.require(lambda self: len(self.model.support) > 0)
    def get_support(self) -> OrderedDict[int, torch.Tensor]:
        """
        Get the support examples for the model.

        Returns
        -------
            OrderedDict[int, torch.Tensor]
            The support examples for the model.
        """
        return self.model.support

    @icontract.require(lambda self: len(self.model.prototypes) > 0)
    def get_prototypes(self) -> torch.Tensor:
        """
        Get the prototypes for the model (the class means of the support embeddings).

        Returns
        -------
        torch.Tensor
            The prototpes for the model.
        """
        return self.model.prototypes

    def save(self, path: str) -> None:
        """
        Save all model parameters to a file.

        Parameters
        ----------
        path : str
            Filename to write the model.

        Returns
        -------
        None
        """
        model_settings = {
            "cov_type": self.cov_type,
            "emb_out_dim": self.emb_out_dim,
            "use_temperature": self.use_temperature,
            "init_temperature": self.temperature.item(),
            "relative_mahal": self.relative_mahal,
        }

        jit_model = torch.jit.script(self.model.embedding_model)
        buffer = io.BytesIO()
        torch.jit.save(jit_model, buffer)
        buffer.seek(0)

        save_data = {
            "embed_jit_save": buffer,
            "feature_names": self.feature_names,
            "label_names": self.label_names,
            "model_head_save": self.model.model_head.state_dict(),
            "outlier_kde": self.outlier_score_kde,
            "settings": model_settings,
            "support": self.model.support,
            "train_summary": self.train_summary,
        }

        torch.save(save_data, path)  # TODO allow model checkpointing

    @classmethod
    def load(cls, path: str) -> Equine:  # noqa: F821
        """
        Load a previously saved EquineProtonet model.

        Parameters
        ----------
        path : str
            The filename of the saved model.

        Returns
        -------
        EquineProtonet
            The reconstituted EquineProtonet object.
        """
        model_save = torch.load(path, weights_only=False)
        support = model_save.get("support")
        jit_model = torch.jit.load(model_save.get("embed_jit_save"))
        eq_model = cls(jit_model, **model_save.get("settings"))

        eq_model.model.model_head.load_state_dict(model_save.get("model_head_save"))
        eq_model.eval()
        eq_model.model.update_support(support)

        eq_model.feature_names = model_save.get("feature_names")
        eq_model.label_names = model_save.get("label_names")
        eq_model.outlier_score_kde = model_save.get("outlier_kde")
        eq_model.train_summary = model_save.get("train_summary")

        return eq_model



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

import os
import torch
from hypothesis import strategies as st
from random import choice
from string import ascii_lowercase, digits

import equine as eq


class BasicEmbeddingModel(torch.nn.Module):
    def __init__(self, tensor_dim, num_classes):
        super(BasicEmbeddingModel, self).__init__()
        self.linear_relu_stack = torch.nn.Sequential(
            torch.nn.Linear(tensor_dim, 32),
            torch.nn.ReLU(),
            torch.nn.Linear(32, 32),
            torch.nn.ReLU(),
            torch.nn.Linear(32, num_classes),
        )

    def forward(self, x):
        logits = self.linear_relu_stack(x)
        return logits


@st.composite
def random_dataset(draw):
    dataset_row_count = draw(st.integers(min_value=100, max_value=100))
    dataset_col_count = draw(st.integers(min_value=1, max_value=1000))
    shape = (dataset_row_count, dataset_col_count)
    dataset_x = torch.rand(shape)

    num_classes = draw(
        st.integers(min_value=3, max_value=int(dataset_row_count / 30))
    )  # requires at least 30 examples per class
    way = draw(st.integers(min_value=2, max_value=num_classes))
    dataset_y = []
    for i in range(num_classes):
        dataset_y += [i] * int(dataset_row_count / num_classes)

    dataset_y += [0] * (dataset_row_count - len(dataset_y))
    dataset_y = torch.Tensor(dataset_y)

    dataset = torch.utils.data.TensorDataset(dataset_x, dataset_y)  # type: ignore

    return dataset, num_classes, way


def use_basic_embedding_model(random_dataset):
    dataset, num_classes, _ = random_dataset
    X, _ = dataset.tensors
    embedding_model = BasicEmbeddingModel(X.shape[1], num_classes)
    return dataset, num_classes, X, embedding_model


def use_save_load_model_tests(model, X, tmp_filename: str = "tmp.eq"):
    old_output = model.predict(X[1:10])
    if os.path.exists(tmp_filename):
        os.remove(tmp_filename)
    model.save(tmp_filename)
    new_model = eq.load_equine_model(tmp_filename)
    new_output = new_model.predict(X[1:10])
    assert (
        torch.nn.functional.mse_loss(old_output.classes, new_output.classes) <= 1e-7
    ), "Class predictions changed on reload"
    assert (
        torch.nn.functional.mse_loss(old_output.ood_scores, new_output.ood_scores)
        <= 1e-7
    ), "OOD predictions changed on reload"

    return new_model, tmp_filename


# return a list of random strings
# based off https://stackoverflow.com/a/34485032
def generate_random_string_list(list_length: int, str_length: int = 3):
    chars = ascii_lowercase + digits
    return [
        "".join(choice(chars) for _ in range(str_length)) for _ in range(list_length)
    ]



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT
import pytest
from conftest import BasicEmbeddingModel

import equine as eq


def test_get_feature_names():
    num_features = 3
    num_classes = 5
    embed_model = BasicEmbeddingModel(num_features, num_classes)

    model = eq.EquineProtonet(embed_model, num_classes)
    assert eq.Equine.get_feature_names(model) is None

    del model.feature_names
    assert eq.Equine.get_feature_names(model) is None

    model = eq.EquineProtonet(embed_model, num_classes, feature_names=["123", "456"])
    assert eq.Equine.get_feature_names(model) == ["123", "456"]


def test_get_label_names():
    num_features = 3
    num_classes = 5
    embed_model = BasicEmbeddingModel(num_features, num_classes)

    model = eq.EquineProtonet(embed_model, num_classes)
    assert eq.Equine.get_label_names(model) is None

    del model.label_names
    assert eq.Equine.get_label_names(model) is None

    model = eq.EquineProtonet(embed_model, num_classes, label_names=["123", "456"])
    assert eq.Equine.get_label_names(model) == ["123", "456"]


def test_validate_feature_label_names():
    num_features = 3
    num_classes = 5
    embed_model = BasicEmbeddingModel(num_features, num_classes)

    # no feature or label names
    model = eq.EquineProtonet(embed_model, num_classes)
    eq.Equine.validate_feature_label_names(
        model, num_features=num_features, num_classes=num_classes
    )

    # with feature names, no label names
    model = eq.EquineProtonet(
        embed_model, num_classes, feature_names=["correct", "num", "features"]
    )
    eq.Equine.validate_feature_label_names(
        model, num_features=num_features, num_classes=num_classes
    )

    # no feature names, with label names
    model = eq.EquineProtonet(
        embed_model, num_classes, label_names=["correct", "num", "of", "label", "names"]
    )
    eq.Equine.validate_feature_label_names(
        model, num_features=num_features, num_classes=num_classes
    )

    # with feature and label names
    model = eq.EquineProtonet(
        embed_model,
        num_classes,
        feature_names=["correct", "num", "features"],
        label_names=["correct", "num", "of", "label", "names"],
    )
    eq.Equine.validate_feature_label_names(
        model, num_features=num_features, num_classes=num_classes
    )


def test_validate_feature_label_names_bad_feature_names():
    num_features = 3
    num_classes = 5
    embed_model = BasicEmbeddingModel(num_features, num_classes)

    model = eq.EquineProtonet(
        embed_model, num_classes, feature_names=["only one feature name"]
    )
    with pytest.raises(ValueError) as exc_info:
        eq.Equine.validate_feature_label_names(
            model, num_features=num_features, num_classes=num_classes
        )
    assert (
        str(exc_info.value)
        == "The length of feature_names (1) does not match the number of data features (3). Update feature_names or set feature_names to None."
    )

    model = eq.EquineProtonet(
        embed_model, num_classes, feature_names=["too", "many", "feature", "names"]
    )
    with pytest.raises(ValueError) as exc_info:
        eq.Equine.validate_feature_label_names(
            model, num_features=num_features, num_classes=num_classes
        )
    assert (
        str(exc_info.value)
        == "The length of feature_names (4) does not match the number of data features (3). Update feature_names or set feature_names to None."
    )


def test_validate_feature_label_names_bad_label_names():
    num_features = 3
    num_classes = 5
    embed_model = BasicEmbeddingModel(num_features, num_classes)

    model = eq.EquineProtonet(
        embed_model, num_classes, label_names=["not", "enough", "label", "names"]
    )
    with pytest.raises(ValueError) as exc_info:
        eq.Equine.validate_feature_label_names(
            model, num_features=num_features, num_classes=num_classes
        )
    assert (
        str(exc_info.value)
        == "The length of label_names (4) does not match the number of classes (5). Update label_names or set label_names to None."
    )

    model = eq.EquineProtonet(
        embed_model,
        num_classes,
        label_names=["too", "many", "label", "names", "to", "pass", "test"],
    )
    with pytest.raises(ValueError) as exc_info:
        eq.Equine.validate_feature_label_names(
            model, num_features=num_features, num_classes=num_classes
        )
    assert (
        str(exc_info.value)
        == "The length of label_names (7) does not match the number of classes (5). Update label_names or set label_names to None."
    )



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

import hypothesis.extra.numpy as hnp
import icontract
import numpy as np
import pytest
import torch
from hypothesis import given, settings, strategies as st

import equine as eq


@st.composite
def support_dataset(draw):
    support_sz = draw(st.integers(min_value=2, max_value=25))
    task_count = draw(st.integers(min_value=2, max_value=10))
    tasks = [x for x in range(task_count)]
    dataset_row_count = draw(
        st.integers(min_value=(support_sz + 1) * task_count, max_value=1000)
    )
    dataset_col_count = draw(st.integers(min_value=1, max_value=1000))
    shape = (dataset_row_count, dataset_col_count)
    dataset_x = torch.rand(shape)
    dataset_y = []
    for task in tasks:
        dataset_y += [task] * (support_sz + 1)

    if len(dataset_x) > len(dataset_y):
        pad = (
            torch.randint(0, task_count, (len(dataset_x) - len(dataset_y), 1))
            .flatten()
            .tolist()
        )
        dataset_y += pad

    dataset_y = torch.Tensor(dataset_y)
    way = draw(st.integers(min_value=2, max_value=task_count))

    return dataset_x, dataset_y, support_sz, tasks, way


@given(dataset=support_dataset())
@settings(deadline=None)
def test_generate_support(dataset):
    train_x, train_y, support_sz, tasks, _ = dataset
    eq.utils.generate_support(train_x, train_y, support_sz, tasks)


@given(dataset=support_dataset())
@settings(deadline=None)
def test_generate_episode(dataset) -> None:
    train_x, train_y, support_sz, tasks, way = dataset
    episode_size = max(len(tasks), train_x.shape[0] // 4)
    eq.utils.generate_episode(train_x, train_y, support_sz, way, episode_size)
    with pytest.raises(ValueError):
        eq.utils.generate_episode(
            train_x, train_y, support_sz, len(tasks) + 1, episode_size
        )


@st.composite
def draw_two_tensors(draw: st.DrawFn) -> tuple[torch.Tensor, torch.Tensor]:
    num_classes = draw(st.integers(min_value=2, max_value=128))
    num_examples = draw(st.integers(min_value=2, max_value=100))
    yh = draw(
        hnp.arrays(
            shape=st.just((num_examples, num_classes)),
            dtype=np.float32,
            elements=st.floats(
                allow_nan=False,
                allow_infinity=False,
                width=32,
                min_value=9.9999998245167e-15,
                max_value=1.0,
            ),
        ).map(lambda arr: torch.tensor(arr, dtype=torch.float32))
    )
    yh = yh / yh.sum(dim=1).unsqueeze(-1)
    true_class = draw(
        st.lists(
            st.integers(min_value=0, max_value=num_classes - 1),
            min_size=num_examples,
            max_size=num_examples,
        ).map(lambda lst: torch.as_tensor(lst, dtype=torch.int64))
    )
    return yh, true_class


@given(draw_two_tensors())
def test_brier_score(two_tensors) -> None:
    yh, yt = two_tensors
    assert eq.utils.brier_score(yh, yt) >= 0.0
    yt = (
        yt.int()
    )  # Regression test to make sure one-hot encoding function doesn't need a LongTensor
    assert eq.utils.brier_score(yh, yt) >= 0.0


@given(draw_two_tensors())
def test_brier_skill_score(two_tensors) -> None:
    yh, yt = two_tensors
    assert eq.utils.brier_skill_score(yh, yt) <= 1.0
    yt = (
        yt.int()
    )  # Regression test to make sure one-hot encoding function doesn't need a LongTensor
    assert eq.utils.brier_skill_score(yh, yt) <= 1.0


@given(draw_two_tensors())
def test_ece(two_tensors) -> None:
    yh, yt = two_tensors
    assert eq.utils.expected_calibration_error(yh, yt) <= 1.0


@given(draw_two_tensors())
def test_metric_summary(two_tensors) -> None:
    yh, yt = two_tensors
    eq_out = eq.EquineOutput(yh, torch.ones(yt.size()), torch.ones(yt.size()))
    metrics = eq.utils.generate_model_metrics(eq_out, yt)
    assert "accuracy" in metrics
    assert "microF1Score" in metrics
    assert "brierScore" in metrics
    assert "brierSkillScore" in metrics
    assert "expectedCalibrationError" in metrics


def test_mahalanobis():
    eps = 10 ** (-4)

    cov = torch.eye(10) * (1 + eps)
    cov = torch.unsqueeze(cov, 0)
    diff = torch.ones((10, 1))

    dist = eq.mahalanobis_distance_nosq(diff, cov)
    target_val = torch.tensor(10 * (1 / (1 + eps)), dtype=dist.dtype)
    assert torch.isclose(dist[0, 0], target_val)

    cov = torch.eye(10) * eps
    cov[0, 0] = cov[0, 0] + 1
    cov = torch.unsqueeze(cov, 0)
    diff = torch.ones((10, 1))

    dist = eq.mahalanobis_distance_nosq(diff, cov)
    target_val = torch.tensor(9 / eps + (1 / (1 + eps)), dtype=dist.dtype)
    assert torch.isclose(dist[0, 0], target_val)

    cov = torch.eye(10) * eps + torch.ones((10, 10))
    cov = torch.unsqueeze(cov, 0)
    diff = torch.ones((10, 1))

    dist = eq.mahalanobis_distance_nosq(diff, cov)
    target_val = torch.tensor(
        (1 / eps) * 10.0 - (100.0) / (eps**2 + eps * 10), dtype=dist.dtype
    )
    assert torch.isclose(dist[0, 0], target_val)


def test_stratified_split_invalid_test_size():
    X = torch.arange(10).float().unsqueeze(1)
    Y = torch.randint(0, 2, (10,))

    with pytest.raises(
        icontract.ViolationError, match="test_size must be between 0 and 1."
    ):
        eq.utils.stratified_train_test_split(X, Y, test_size=1.5)

    with pytest.raises(
        icontract.ViolationError, match="test_size must be between 0 and 1."
    ):
        eq.utils.stratified_train_test_split(X, Y, test_size=0)


def test_stratified_split_mismatched_lengths():
    X = torch.arange(10).float().unsqueeze(1)
    Y = torch.randint(0, 2, (8,))  # Mismatched length

    with pytest.raises(
        icontract.ViolationError, match="X and Y must have the same number of samples."
    ):
        eq.utils.stratified_train_test_split(X, Y, test_size=0.3)


def test_stratified_split_imbalanced_classes():
    X = torch.arange(50).float().unsqueeze(1)
    Y = torch.cat(
        [torch.zeros(45, dtype=torch.long), torch.ones(5, dtype=torch.long)]
    )  # Imbalanced

    _, _, train_y, test_y = eq.utils.stratified_train_test_split(X, Y, test_size=0.2)

    # Check that minority class is present in both splits
    assert (train_y == 1).sum().item() > 0
    assert (test_y == 1).sum().item() > 0

    # Verify counts
    total_counts = torch.tensor([(Y == cls).sum().item() for cls in torch.unique(Y)])
    expected_test_counts = (total_counts.float() * 0.2).round().long()
    expected_train_counts = total_counts - expected_test_counts

    actual_train_counts = torch.tensor(
        [(train_y == cls).sum().item() for cls in torch.unique(Y)]
    )
    actual_test_counts = torch.tensor(
        [(test_y == cls).sum().item() for cls in torch.unique(Y)]
    )

    assert torch.all(actual_train_counts == expected_train_counts)
    assert torch.all(actual_test_counts == expected_test_counts)



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from typing import Any, Optional, TypeVar

import icontract
import torch
from abc import ABC, abstractmethod
from collections import OrderedDict
from torch.utils.data import TensorDataset

from .equine_output import EquineOutput

# A type variable for Equine objects
AnyEquine = TypeVar("AnyEquine", bound="Equine")


class Equine(torch.nn.Module, ABC):
    """EQUINE Abstract Base Class (ABC):
    EQUINE is set up to extend torch's nn.Module to enrich it with
    a method that enables uncertainty quantification and visualization. Most
    importantly, the `.predict()` method must be outfitted to return an
    EquineOutput object that contains both the class probabilities
    *and* an out-of-distribution (ood) score.

    Parameters
    ----------
    embedding_model : torch.nn.Module
        The embedding model to use.
    head_layers : int, optional
        The number of layers to use in the model head, by default 1.
    device : str, optional
        The device to train the equine model on (defaults to cpu).
    feature_names : list[str], optional
        List of strings of the names of the tabular features (ex ["duration", "fiat_mean", ...])
    label_names : list[str], optional
        List of strings of the names of the labels (ex ["streaming", "voip", ...])

    Attributes
    ----------
    device : str
        The device to train the equine model on (defaults to cpu).
    embedding_model : torch.nn.Module
        The neural embedding model to enrich with uncertainty quantification.
    feature_names : list[str], optional
        List of strings of the names of the tabular features (ex ["duration", "fiat_mean", ...])
    head_layers : int
        The number of linear layers to append to the embedding model (default 1, not always used).
    label_names : list[str], optional
        List of strings of the names of the labels (ex ["streaming", "voip", ...])
    train_summary : dict[str, Any]
        A dictionary containing information about the model training.

    Raises
    ------
    NotImplementedError
        If any of the abstract methods are not implemented.
    """

    def __init__(
        self,
        embedding_model: torch.nn.Module,
        head_layers: int = 1,
        device: str = "cpu",
        feature_names: Optional[list[str]] = None,
        label_names: Optional[list[str]] = None,
    ) -> None:
        super().__init__()
        self.embedding_model = embedding_model
        self.head_layers = head_layers
        self.train_summary: dict[str, Any] = {
            "numTrainExamples": 0,
            "dateTrained": "",
            "modelType": "",
        }
        self.device = device
        self.to(device)
        self.embedding_model.to(device)
        self.feature_names = feature_names
        self.label_names = label_names

        self.support: OrderedDict[int, torch.Tensor] = OrderedDict()
        self.support_embeddings: OrderedDict[int, torch.Tensor] = OrderedDict()
        self.prototypes: torch.Tensor = torch.Tensor()

    @abstractmethod
    def forward(self, X: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the model. This is to preserve the usual behavior
        of torch.nn.Module.

        Parameters
        ----------
        X : torch.Tensor
            The input data.

        Returns
        -------
        torch.Tensor
            The output of the model.

        """
        raise NotImplementedError

    @abstractmethod
    def predict(self, X: torch.Tensor) -> EquineOutput:
        """
        Upon implementation, predicts the class logits and out-of-distribution (ood) scores for the
        given input data.

        Parameters
        ----------
        X : torch.Tensor
            The input data.

        Returns
        -------
        EquineOutput
            An EquineOutput object containing the class probabilities and OOD scores.
        """
        raise NotImplementedError

    @abstractmethod
    def train_model(
        self, dataset: TensorDataset, *args: Any, **kwargs: Any
    ) -> dict[str, Any]:
        """
        Upon implementation, train the model on the given dataset.

        Parameters
        ----------
        dataset : TensorDataset
            TensorDataset containing the training data.
        **kwargs
            Additional keyword arguments to pass to the training function.

        Returns
        -------
        dict[str, Any]
            Dictionary containing summary training information and any other data
            Note that at least one key should be 'train_summary'
        """
        raise NotImplementedError

    @abstractmethod
    def get_prototypes(self) -> torch.Tensor:
        """
        Upon implementation, returns the prototype embeddings

        Returns
        -------
        torch.Tensor
            A torch tensor of the prototype embeddings
        """
        raise NotImplementedError

    @abstractmethod
    def save(self, path: str) -> None:
        """
        Upon implementation, save the model to the given file path.

        Parameters
        ----------
        path : str
            File path to save the model to.
        """
        raise NotImplementedError

    @classmethod  # type: ignore
    def load(cls: AnyEquine, path: str) -> AnyEquine:  # noqa: F821 # type: ignore
        """
        Upon implementation, load the model from the given file path.

        Parameters
        ----------
        path : str
            File path to load the model from.

        Returns
        -------
        Equine
            Loaded model object.
        """
        raise NotImplementedError

    def get_label_names(self) -> Optional[list[str]]:
        """
        Retrieve the label names used in the model.

        Returns
        -------
        Optional[list[str]]
            A list of label names if available; otherwise, None.
        """
        if hasattr(self, "label_names"):
            return self.label_names
        return None

    def get_feature_names(self) -> Optional[list[str]]:
        """
        Retrieve the feature names used in the model.

        Returns
        -------
        Optional[list[str]]
            A list of feature names if available; otherwise, None.
        """
        if hasattr(self, "feature_names"):
            return self.feature_names
        return None

    @icontract.require(
        lambda num_features, num_classes: num_features > 0 and num_classes > 0
    )
    def validate_feature_label_names(self, num_features: int, num_classes: int) -> None:
        """
        Validate that the feature names and label names, if provided, match the expected counts.

        Parameters
        ----------
        num_features : int
            The expected number of features.
        num_classes : int
            The expected number of classes.

        Raises
        ------
        ValueError
            If the length of `feature_names` does not match `num_features`, or
            if the length of `label_names` does not match `num_classes`.
        """
        feature_names = self.get_feature_names()
        if feature_names is not None and len(feature_names) != num_features:
            raise ValueError(
                f"The length of feature_names ({len(feature_names)}) does not match the number of data features ({num_features}). Update feature_names or set feature_names to None."
            )

        label_names = self.get_label_names()
        if label_names is not None and len(label_names) != num_classes:
            raise ValueError(
                f"The length of label_names ({len(label_names)}) does not match the number of classes ({num_classes}). Update label_names or set label_names to None."
            )



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from typing import Any, Optional, Union

import icontract
import io
import math
import torch
from beartype import beartype
from collections import OrderedDict
from collections.abc import Callable
from datetime import datetime
from torch.utils.data import DataLoader, TensorDataset
from tqdm import tqdm

from .equine import Equine, EquineOutput
from .utils import generate_support, generate_train_summary, stratified_train_test_split

BatchType = tuple[torch.Tensor, ...]
# -------------------------------------------------------------------------------
# Note that the below code for
# * `_random_ortho`,
# * `_RandomFourierFeatures``, and
# * `_Laplace`
# is copied and modified from https://github.com/y0ast/DUE/blob/main/due/sngp.py
# under its original MIT license, redisplayed here:
# -------------------------------------------------------------------------------
# MIT License
#
# Copyright (c) 2021 Joost van Amersfoort
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ------------------------------------------------------------------------------
# Following the recommendation of their README at https://github.com/y0ast/DUE
# we encourage anyone using this code in their research to cite the following papers:
#
# @article{van2021on,
#   title={On Feature Collapse and Deep Kernel Learning for Single Forward Pass Uncertainty},
#   author={van Amersfoort, Joost and Smith, Lewis and Jesson, Andrew and Key, Oscar and Gal, Yarin},
#   journal={arXiv preprint arXiv:2102.11409},
#   year={2021}
# }
#
# @article{liu2020simple,
#  title={Simple and principled uncertainty estimation with deterministic deep learning via distance awareness},
#  author={Liu, Jeremiah and Lin, Zi and Padhy, Shreyas and Tran, Dustin and Bedrax Weiss, Tania and Lakshminarayanan, Balaji},
#  journal={Advances in Neural Information Processing Systems},
#  volume={33},
#  pages={7498--7512},
#  year={2020}
# }


@beartype
def _random_ortho(n: int, m: int) -> torch.Tensor:
    """
     Generate a random orthonormal matrix.

     Parameters
     ----------
     n : int
         The number of rows.
    m : int
         The number of columns.

     Returns
     -------
     torch.Tensor
         The random orthonormal matrix.
    """
    q, _ = torch.linalg.qr(torch.randn(n, m))
    return q


@beartype
class _RandomFourierFeatures(torch.nn.Module):
    """
    A private class to generate random Fourier features for the embedding model.
    """

    def __init__(
        self, in_dim: int, num_random_features: int, feature_scale: Optional[float]
    ) -> None:
        """
        Initialize the _RandomFourierFeatures module, which generates random Fourier features
        for the embedding model.

        Parameters
        ----------
        in_dim : int
            The input dimensionality.
        num_random_features : int
            The number of random Fourier features to generate.
        feature_scale : Optional[float]
            The scaling factor for the random Fourier features. If None, defaults to sqrt(num_random_features / 2).
        """
        super().__init__()
        if feature_scale is None:
            feature_scale = math.sqrt(num_random_features / 2)

        self.register_buffer("feature_scale", torch.tensor(feature_scale))

        if num_random_features <= in_dim:
            W = _random_ortho(in_dim, num_random_features)
        else:
            # generate blocks of orthonormal rows which are not necessarily orthonormal
            # to each other.
            dim_left = num_random_features
            ws = []
            while dim_left > in_dim:
                ws.append(_random_ortho(in_dim, in_dim))
                dim_left -= in_dim
            ws.append(_random_ortho(in_dim, dim_left))
            W = torch.cat(ws, 1)

        feature_norm = torch.randn(W.shape) ** 2
        W = W * feature_norm.sum(0).sqrt()
        self.register_buffer("W", W)

        b = torch.empty(num_random_features).uniform_(0, 2 * math.pi)
        self.register_buffer("b", b)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Compute the forward pass of the _RandomFourierFeatures module.

        Parameters
        ----------
        x : torch.Tensor
            The input tensor of shape (batch_size, in_dim).

        Returns
        -------
        torch.Tensor
            The output tensor of shape (batch_size, num_random_features).
        """
        k = torch.cos(x @ self.W + self.b)
        k = k / self.feature_scale

        return k


class _Laplace(torch.nn.Module):
    """
    A private class to compute a Laplace approximation to a Gaussian Process (GP)
    """

    def __init__(
        self,
        feature_extractor: torch.nn.Module,
        num_deep_features: int,
        num_gp_features: int,
        normalize_gp_features: bool,
        num_random_features: int,
        num_outputs: int,
        feature_scale: Optional[float],
        mean_field_factor: Optional[float],  # required for classification problems
        ridge_penalty: float = 1.0,
    ) -> None:
        """
        Initialize the _Laplace module.

        Parameters
        ----------
        feature_extractor : torch.nn.Module
            The feature extractor module.
        num_deep_features : int
            The number of features output by the feature extractor.
        num_gp_features : int
            The number of features to use in the Gaussian process.
        normalize_gp_features : bool
            Whether to normalize the GP features.
        num_random_features : int
            The number of random Fourier features to use.
        num_outputs : int
            The number of outputs of the model.
        feature_scale : Optional[float]
            The scaling factor for the random Fourier features.
        mean_field_factor : Optional[float]
            The mean-field factor for the Gaussian-Softmax approximation.
            Required for classification problems.
        ridge_penalty : float, optional
            The ridge penalty for the Laplace approximation.
        """
        super().__init__()
        self.feature_extractor = feature_extractor
        self.mean_field_factor = mean_field_factor
        self.ridge_penalty = ridge_penalty
        self.train_batch_size = 0  # to be set later

        if num_gp_features > 0:
            self.num_gp_features = num_gp_features
            self.register_buffer(
                "random_matrix",
                torch.normal(0, 0.05, (num_gp_features, num_deep_features)),
            )
            self.jl: Callable = lambda x: torch.nn.functional.linear(
                x, self.random_matrix
            )
        else:
            self.num_gp_features: int = num_deep_features
            self.jl: Callable = lambda x: x  # Identity

        self.normalize_gp_features = normalize_gp_features
        if normalize_gp_features:
            self.normalize: torch.nn.LayerNorm = torch.nn.LayerNorm(num_gp_features)

        self.rff: _RandomFourierFeatures = _RandomFourierFeatures(
            num_gp_features, num_random_features, feature_scale
        )
        self.beta: torch.nn.Linear = torch.nn.Linear(num_random_features, num_outputs)

        self.num_data = 0  # to be set later
        self.register_buffer("seen_data", torch.tensor(0))

        precision = torch.eye(num_random_features) * self.ridge_penalty
        self.register_buffer("precision", precision)

        self.recompute_covariance = True
        self.register_buffer("covariance", torch.eye(num_random_features))
        self.training_parameters_set = False

    def reset_precision_matrix(self) -> None:
        """
        Reset the precision matrix to the identity matrix times the ridge penalty.
        """
        identity = torch.eye(self.precision.shape[0], device=self.precision.device)
        self.precision: torch.Tensor = identity * self.ridge_penalty
        self.seen_data: torch.Tensor = torch.tensor(0)
        self.recompute_covariance = True

    @icontract.require(lambda num_data: num_data > 0)
    @icontract.require(
        lambda num_data, batch_size: (0 < batch_size) & (batch_size <= num_data)
    )
    def set_training_params(self, num_data: int, batch_size: int) -> None:
        """
        Set the training parameters for the Laplace approximation.

        Parameters
        ----------
        num_data : int
            The total number of data points.
        batch_size : int
            The batch size to use during training.
        """
        self.num_data: int = num_data
        self.train_batch_size: int = batch_size
        self.training_parameters_set: bool = True

    @icontract.require(lambda self: self.mean_field_factor is not None)
    def mean_field_logits(
        self, logits: torch.Tensor, pred_cov: torch.Tensor
    ) -> torch.Tensor:
        """
        Compute the mean-field logits for the Gaussian-Softmax approximation.

        Parameters
        ----------
        logits : torch.Tensor
            The logits tensor of shape (batch_size, num_outputs).
        pred_cov : torch.Tensor
            The predicted covariance matrix of shape (batch_size, batch_size).

        Returns
        -------
        torch.Tensor
            The mean-field logits tensor of shape (batch_size, num_outputs).
        """
        # Mean-Field approximation as alternative to MC integration of Gaussian-Softmax
        # Based on: https://arxiv.org/abs/2006.07584

        logits_scale = torch.sqrt(1.0 + torch.diag(pred_cov) * self.mean_field_factor)
        if self.mean_field_factor > 0:  # type: ignore
            logits = logits / logits_scale.unsqueeze(-1)

        return logits

    @icontract.require(lambda self: self.training_parameters_set)
    def forward(
        self, x: torch.Tensor
    ) -> Union[torch.Tensor, tuple[torch.Tensor, torch.Tensor]]:
        """
        Compute the forward pass of the Laplace approximation to the Gaussian Process.

        Parameters
        ----------
        x : torch.Tensor
            The input tensor of shape (batch_size, num_features).

        Returns
        -------
        Union[torch.Tensor, tuple[torch.Tensor, torch.Tensor]]
            If the model is in training mode, returns the predicted mean of shape (batch_size, 1).
            If the model is in evaluation mode, returns a tuple containing the predicted mean of shape (batch_size, 1)
            and the predicted covariance matrix of shape (batch_size, batch_size).
        """
        f = self.feature_extractor(x)
        f_reduc = self.jl(f)
        if self.normalize_gp_features:
            f_reduc = self.normalize(f_reduc)

        k = self.rff(f_reduc)

        pred = self.beta(k)

        if self.training:
            precision_minibatch = k.t() @ k
            self.precision += precision_minibatch
            self.seen_data += x.shape[0]

            assert (
                self.seen_data <= self.num_data
            ), "Did not reset precision matrix at start of epoch"
        else:
            assert self.seen_data > (
                self.num_data - self.train_batch_size
            ), "Not seen sufficient data for precision matrix"

            if self.recompute_covariance:
                with torch.no_grad():
                    eps = 1e-7
                    jitter = eps * torch.eye(
                        self.precision.shape[1],
                        device=self.precision.device,
                    )
                    u, info = torch.linalg.cholesky_ex(self.precision + jitter)
                    assert (info == 0).all(), "Precision matrix inversion failed!"
                    torch.cholesky_inverse(u, out=self.covariance)

                self.recompute_covariance: bool = False

            with torch.no_grad():
                pred_cov = k @ ((self.covariance @ k.t()) * self.ridge_penalty)

            if self.mean_field_factor is None:
                return pred, pred_cov
            else:
                pred = self.mean_field_logits(pred, pred_cov)

        return pred


# -------------------------------------------------------------------------------
# EquineGP, below, demonstrates how to adapt that approach in EQUINE
@beartype
class EquineGP(Equine):
    """
    An example of an EQUINE model that builds upon the approach in "Spectral Norm
    Gaussian Processes" (SNGP). This wraps any pytorch embedding neural network and provides
    the `forward`, `predict`, `save`, and `load` methods required by Equine.

    Notes
    -----
    Although this model build upon the approach in SNGP, it does not enforce the spectral normalization
    and ResNet architecture required for SNGP. Instead, it is a simple wrapper around
    any pytorch embedding neural network. Your mileage may vary.
    """

    def __init__(
        self,
        embedding_model: torch.nn.Module,
        emb_out_dim: int,
        num_classes: int,
        use_temperature: bool = False,
        init_temperature: float = 1.0,
        device: str = "cpu",
        feature_names: Optional[list[str]] = None,
        label_names: Optional[list[str]] = None,
    ) -> None:
        """
        Initialize the EquineGP model.

        Parameters
        ----------
        embedding_model : torch.nn.Module
            Neural Network feature embedding.
        emb_out_dim : int
            The number of deep features from the feature embedding.
        num_classes : int
            The number of output classes this model predicts.
        use_temperature : bool, optional
            Whether to use temperature scaling after training.
        init_temperature : float, optional
            What to use as the initial temperature (1.0 has no effect).
        device : str, optional
            Either 'cuda' or 'cpu'.
        feature_names : list[str], optional
            List of strings of the names of the tabular features (ex ["duration", "fiat_mean", ...])
        label_names : list[str], optional
            List of strings of the names of the labels (ex ["streaming", "voip", ...])
        """
        super().__init__(
            embedding_model, feature_names=feature_names, label_names=label_names
        )
        self.num_deep_features = emb_out_dim
        self.num_gp_features = emb_out_dim
        self.normalize_gp_features = True
        self.num_random_features = 1024
        self.num_outputs = num_classes
        self.mean_field_factor = 25
        self.ridge_penalty = 1
        self.feature_scale: float = 2.0
        self.use_temperature = use_temperature
        self.init_temperature = init_temperature
        self.register_buffer(
            "temperature", torch.Tensor(self.init_temperature * torch.ones(1))
        )
        self.model: _Laplace = _Laplace(
            self.embedding_model,
            self.num_deep_features,
            self.num_gp_features,
            self.normalize_gp_features,
            self.num_random_features,
            self.num_outputs,
            self.feature_scale,
            self.mean_field_factor,
            self.ridge_penalty,
        )
        self.device_type = device
        self.device: torch.device = torch.device(self.device_type)
        self.model.to(self.device)

    def train_model(
        self,
        dataset: TensorDataset,
        loss_fn: Callable,
        opt: torch.optim.Optimizer,
        num_epochs: int,
        batch_size: int = 64,
        calib_frac: float = 0.1,
        num_calibration_epochs: int = 2,
        calibration_lr: float = 0.01,
        vis_support: bool = False,
        support_size: int = 25,
    ) -> dict[str, Any]:
        """
        Train or fine-tune an EquineGP model.

        Parameters
        ----------
        dataset : TensorDataset
            An iterable, pytorch TensorDataset.
        loss_fn : Callable
            A pytorch loss function, e.g., torch.nn.CrossEntropyLoss().
        opt : torch.optim.Optimizer
            A pytorch optimizer, e.g., torch.optim.Adam().
        num_epochs : int
            The desired number of epochs to use for training.
        batch_size : int, optional
            The number of samples to use per batch.
        calib_frac : float, optional
            Fraction of training data to use in temperature scaling.
        num_calibration_epochs : int, optional
            The desired number of epochs to use for temperature scaling.
        calibration_lr : float, optional
            Learning rate for temperature scaling.

        Returns
        -------
        dict[str, Any]
            A dict containing a dict of summary stats and a dataloader for the calibration data.

        Notes
        -------
        - If `use_temperature` is True, temperature scaling will be used after training.
        - The calibration data is used to calibrate the temperature scaling.
        """
        X, Y = dataset[:]
        calib_x = torch.Tensor()
        calib_y = torch.Tensor()
        if self.use_temperature:
            train_x, calib_x, train_y, calib_y = stratified_train_test_split(
                X, Y, test_size=calib_frac
            )
            dataset = TensorDataset(torch.Tensor(train_x), torch.Tensor(train_y))
            self.temperature: torch.Tensor = torch.Tensor(
                self.init_temperature * torch.ones(1)
            ).type_as(self.temperature)

        self.validate_feature_label_names(X.shape[-1], self.num_outputs)

        train_loader = DataLoader(
            dataset, batch_size=batch_size, shuffle=True, drop_last=True
        )
        self.model.set_training_params(len(dataset), batch_size)
        self.model.train()
        for _ in tqdm(range(num_epochs)):
            self.model.reset_precision_matrix()
            epoch_loss = 0.0
            for i, (xs, labels) in enumerate(train_loader):
                opt.zero_grad()
                xs = xs.to(self.device)
                labels = labels.to(self.device)
                yhats = self.model(xs)
                loss = loss_fn(yhats, labels.to(torch.long))
                loss.backward()
                opt.step()
                epoch_loss += loss.item()
        self.model.eval()

        if vis_support:
            self.update_support(dataset.tensors[0], dataset.tensors[1], support_size)

        calibration_loader = None
        if self.use_temperature:
            dataset_calibration = TensorDataset(
                torch.Tensor(calib_x), torch.Tensor(calib_y)
            )
            calibration_loader = DataLoader(
                dataset_calibration,
                batch_size=batch_size,
                shuffle=True,
                drop_last=False,
            )
            self.calibrate_temperature(
                calibration_loader, num_calibration_epochs, calibration_lr
            )

        _, train_y = dataset[:]
        date_trained = datetime.now().strftime("%m/%d/%Y, %H:%M:%S")
        self.train_summary: dict[str, Any] = generate_train_summary(
            self, train_y, date_trained
        )

        return_dict: dict[str, Any] = dict()
        return_dict["train_summary"] = self.train_summary
        return_dict["calibration_loader"] = calibration_loader

        return return_dict

    def update_support(
        self, support_x: torch.Tensor, support_y: torch.Tensor, support_size: int
    ) -> None:
        """Function to update protonet support examples with given examples.

        Parameters
        ----------
        support_x : torch.Tensor
            Tensor containing support examples for protonet.
        support_y : torch.Tensor
            Tensor containing labels for given support examples.

        Returns
        -------
        None
        """

        labels, counts = torch.unique(support_y, return_counts=True)
        support = OrderedDict()
        for label, count in list(zip(labels.tolist(), counts.tolist())):
            class_support = generate_support(
                support_x,
                support_y,
                support_size=min(count, support_size),
                selected_labels=[label],
            )
            support.update(class_support)

        self.support = support

        support_embeddings = OrderedDict().fromkeys(self.support.keys(), torch.Tensor())
        for label in support:
            support_embeddings[label] = self.compute_embeddings(support[label])

        self.support_embeddings = support_embeddings
        self.prototypes: torch.Tensor = self.compute_prototypes()

    def compute_embeddings(self, x: torch.Tensor) -> torch.Tensor:
        """
        Method for computing deep embeddings for given input tensor.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor for generating embeddings.

        Returns
        -------
        torch.Tensor
            Output embeddings .
        """
        f = self.model.feature_extractor(x)
        f_reduc = self.model.jl(f)
        if self.model.normalize_gp_features:
            f_reduc = self.model.normalize(f_reduc)

        return self.model.rff(f_reduc)

    @icontract.require(lambda self: len(self.support) > 0)
    def compute_prototypes(self) -> torch.Tensor:
        """
        Method for computing class prototypes based on given support examples.
        ``Prototypes'' in this context are the means of the support embeddings for each class.

        Returns
        -------
        torch.Tensor
            Tensors of prototypes for each of the given classes in the support.
        """
        # Compute support embeddings
        support_embeddings = OrderedDict().fromkeys(self.support.keys())
        for label in self.support:
            support_embeddings[label] = self.compute_embeddings(self.support[label])

        # Compute prototype for each class
        proto_list = []
        for label in self.support:  # look at doing functorch
            class_prototype = torch.mean(support_embeddings[label], dim=0)  # type: ignore
            proto_list.append(class_prototype)

        prototypes = torch.stack(proto_list)

        return prototypes

    @icontract.require(lambda self: len(self.support) > 0)
    def get_support(self) -> OrderedDict[int, torch.Tensor]:
        """
        Method for returning support examples used in training.

        Returns
        -------
            OrderedDict[int, torch.Tensor]
            Dictionary containing support examples for each class.
        """
        return self.support

    @icontract.require(lambda self: len(self.prototypes) > 0)
    def get_prototypes(self) -> torch.Tensor:
        """
        Method for returning class prototypes.

        Returns
        -------
        torch.Tensor
            Tensors of prototypes for each of the given classes in the support.
        """
        return self.prototypes

    @icontract.require(lambda num_calibration_epochs: 0 < num_calibration_epochs)
    @icontract.require(lambda calibration_lr: calibration_lr > 0.0)
    def calibrate_temperature(
        self,
        calibration_loader: DataLoader[BatchType],
        num_calibration_epochs: int = 1,
        calibration_lr: float = 0.01,
    ) -> None:
        """
        Fine-tune the temperature after training. Note this function is also run at the conclusion of train_model.

        Parameters
        ----------
        calibration_loader : DataLoader
            Data loader returned by train_model.
        num_calibration_epochs : int, optional
            Number of epochs to tune temperature.
        calibration_lr : float, optional
            Learning rate for temperature optimization.
        """
        self.temperature.requires_grad = True
        loss_fn = torch.nn.functional.cross_entropy
        optimizer = torch.optim.Adam([self.temperature], lr=calibration_lr)
        for _ in range(num_calibration_epochs):
            for xs, labels in calibration_loader:
                optimizer.zero_grad()
                xs = xs.to(self.device)
                labels = labels.to(self.device)
                with torch.no_grad():
                    logits = self.model(xs)
                logits = logits / self.temperature
                loss = loss_fn(logits, labels.to(torch.long))
                loss.backward()
                optimizer.step()
        self.temperature.requires_grad = False

    def forward(self, X: torch.Tensor) -> torch.Tensor:
        """
        EquineGP forward function, generates logits for classification.

        Parameters
        ----------
        X : torch.Tensor
            Input tensor for generating predictions.

        Returns
        -------
        torch.Tensor
            Output probabilities computed.
        """
        X = X.to(self.device)
        preds = self.model(X)
        return preds / self.temperature.to(self.device)

    @icontract.ensure(
        lambda result: all((0 <= result.ood_scores) & (result.ood_scores <= 1.0))
    )
    def predict(self, X: torch.Tensor) -> EquineOutput:
        """
        Predict function for EquineGP, inherited and implemented from Equine.

        Parameters
        ----------
        X : torch.Tensor
            Input tensor.

        Returns
        -------
        EquineOutput
            Output object containing prediction probabilities and OOD scores.
        """
        logits = self(X)
        preds = torch.softmax(logits, dim=1)
        equiprobable = torch.ones(self.num_outputs) / self.num_outputs
        max_entropy = torch.sum(torch.special.entr(equiprobable))
        ood_score = torch.sum(torch.special.entr(preds), dim=1) / max_entropy
        embeddings = self.compute_embeddings(X)
        eq_out = EquineOutput(
            classes=preds, ood_scores=ood_score, embeddings=embeddings
        )  # TODO return embeddings

        self.validate_feature_label_names(X.shape[-1], self.num_outputs)

        return eq_out

    def save(self, path: str) -> None:
        """
        Function to save all model parameters to a file.

        Parameters
        ----------
        path : str
            Filename to write the model.
        """
        model_settings = {
            "emb_out_dim": self.num_deep_features,
            "num_classes": self.num_outputs,
            "use_temperature": self.use_temperature,
            "init_temperature": self.temperature.item(),
            "device": self.device_type,
        }

        jit_model = torch.jit.script(self.model.feature_extractor)
        buffer = io.BytesIO()
        torch.jit.save(jit_model, buffer)
        buffer.seek(0)

        laplace_sd = self.model.state_dict()
        keys_to_delete = []
        for key in laplace_sd:
            if "feature_extractor" in key:
                keys_to_delete.append(key)
        for key in keys_to_delete:
            del laplace_sd[key]

        save_data = {
            "embed_jit_save": buffer,
            "feature_names": self.feature_names,
            "label_names": self.label_names,
            "laplace_model_save": laplace_sd,
            "num_data": self.model.num_data,
            "settings": model_settings,
            "support": self.support,
            "train_batch_size": self.model.train_batch_size,
            "train_summary": self.train_summary,
        }

        torch.save(save_data, path)  # TODO allow model checkpointing

    @classmethod
    def load(cls, path: str) -> Equine:
        """
        Function to load previously saved EquineGP model.

        Parameters
        ----------
        path : str
            Input filename.

        Returns
        -------
        EquineGP
            The reconstituted EquineGP object.
        """
        model_save = torch.load(path, weights_only=False)
        jit_model = torch.jit.load(model_save.get("embed_jit_save"))
        eq_model = cls(jit_model, **model_save.get("settings"))

        eq_model.feature_names = model_save.get("feature_names")
        eq_model.label_names = model_save.get("label_names")
        eq_model.train_summary = model_save.get("train_summary")

        eq_model.model.load_state_dict(
            model_save.get("laplace_model_save"), strict=False
        )
        eq_model.model.seen_data = model_save.get("laplace_model_save").get("seen_data")

        eq_model.model.set_training_params(
            model_save.get("num_data"), model_save.get("train_batch_size")
        )
        eq_model.eval()

        support = model_save.get("support")
        if len(support) > 0:
            eq_model.support = support
            eq_model.prototypes = eq_model.compute_prototypes()

        return eq_model



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT
import os
import pytest
import torch
from conftest import (
    BasicEmbeddingModel,
    generate_random_string_list,
    random_dataset,
    use_basic_embedding_model,
    use_save_load_model_tests,
)
from hypothesis import given, settings, strategies as st

import equine as eq


@given(
    data_shape=st.tuples(
        st.integers(min_value=1, max_value=1000),
        st.integers(min_value=1, max_value=1000),
    ),
    num_classes=st.integers(min_value=2, max_value=256),
)
@settings(deadline=None)
def test_compute_embeddings(data_shape, num_classes):
    queries = torch.rand(data_shape)
    embed_model = BasicEmbeddingModel(data_shape[1], num_classes)
    model = eq.EquineProtonet(embed_model, num_classes)
    model.model.compute_embeddings(queries)


@st.composite
def embeddings(draw):
    dataset_row_count = draw(st.integers(min_value=100, max_value=1000))
    dataset_col_count = draw(st.integers(min_value=3, max_value=10))
    shape = (dataset_row_count, dataset_col_count)
    support_embeddings = torch.rand(shape)
    query_embeddings = torch.rand(shape)
    ptrs = [0]
    curr_idx = 0

    for i in range(dataset_col_count - 1):
        curr_idx += int(dataset_row_count / dataset_col_count)
        ptrs.append(curr_idx)

    ptrs.append(dataset_row_count - 1)
    ptrs = torch.Tensor(ptrs).to(torch.long)

    return support_embeddings, ptrs, query_embeddings


@given(random_dataset=random_dataset())
@settings(deadline=None)
def test_train_episodes(random_dataset):
    dataset, num_classes, way = random_dataset
    num_shot = 3
    num_episodes = 10
    episode_size = 512

    X, Y = dataset.tensors
    num_deep_features = 32
    embed_model = BasicEmbeddingModel(X.shape[1], num_deep_features)
    model = eq.EquineProtonet(embed_model, num_deep_features)
    model.train_model(
        dataset,
        way=way,
        support_size=num_shot,
        num_episodes=num_episodes,
        episode_size=episode_size,
    )

    assert model.model.training is False, "Model leaves training mode"
    assert len(model.model.support) == num_classes  # type: ignore
    # Test on multiple predictions
    eq_out = model.predict(X)
    assert len(eq_out.classes) == len(X)
    assert len(eq_out.ood_scores) == len(X)
    # Test on single prediction
    pred_out = model(X[0])
    assert len(pred_out) == 1, "Single prediction works"
    eq_out = model.predict(X[0])
    assert len(eq_out.classes) == 1, "Single prediction works"

    support = model.get_support()
    assert (
        support is not None and len(support) == num_classes
    ), "Support set is correct size"
    prototypes = model.get_prototypes()
    assert (
        prototypes is not None and len(prototypes) == num_classes
    ), "Prototypes set is correct size"
    assert (
        model.model.support is not None and len(model.model.support) == num_classes
    ), "Support set is correct size"


@given(random_dataset=random_dataset())
@settings(deadline=None)
def test_train_episodes_shared_reg(random_dataset):
    dataset, num_classes, way = random_dataset
    num_shot = 3
    num_episodes = 10
    episode_size = 512

    X, Y = dataset.tensors
    num_deep_features = 32
    embed_model = BasicEmbeddingModel(X.shape[1], num_deep_features)
    model = eq.EquineProtonet(
        embed_model, num_deep_features, cov_type=eq.CovType.DIAGONAL
    )
    model.cov_reg_type = "shared"
    model.model.cov_reg_type = "shared"
    model.train_model(
        dataset,
        way=way,
        support_size=num_shot,
        num_episodes=num_episodes,
        episode_size=episode_size,
    )

    assert model.model.training is False, "Model leaves training mode"
    assert len(model.model.support) == num_classes  # type: ignore
    # Test on multiple predictions
    eq_out = model.predict(X)
    assert len(eq_out.classes) == len(X)
    assert len(eq_out.ood_scores) == len(X)
    # Test on single prediction
    pred_out = model(X[0])
    assert len(pred_out) == 1, "Single prediction works"
    eq_out = model.predict(X[0])
    assert len(eq_out.classes) == 1, "Single prediction works"

    support = model.get_support()
    assert (
        support is not None and len(support) == num_classes
    ), "Support set is correct size"
    model.update_support(X, Y, 0.5)
    assert (
        model.model.support is not None and len(model.model.support) == num_classes
    ), "Support set is correct size"


@given(random_dataset=random_dataset())
@settings(deadline=None)
def test_train_episodes_full_cov(random_dataset):
    dataset, num_classes, way = random_dataset
    num_shot = 20
    num_episodes = 5
    episode_size = 512

    X, Y = dataset.tensors
    num_deep_features = 4
    embed_model = BasicEmbeddingModel(X.shape[1], num_deep_features)
    model = eq.EquineProtonet(embed_model, num_deep_features, cov_type=eq.CovType.FULL)
    model.cov_reg_type = "epsilon"
    model.model.cov_reg_type = "epsilon"
    model.train_model(
        dataset,
        way=way,
        support_size=num_shot,
        num_episodes=num_episodes,
        episode_size=episode_size,
    )

    assert model.model.training is False, "Model leaves training mode"
    assert len(model.model.support) == num_classes  # type: ignore
    # Test on multiple predictions
    eq_out = model.predict(X)
    assert len(eq_out.classes) == len(X)
    assert len(eq_out.ood_scores) == len(X)
    # Test on single prediction
    pred_out = model(X[0])
    assert len(pred_out) == 1, "Single prediction works"
    eq_out = model.predict(X[0])
    assert len(eq_out.classes) == 1, "Single prediction works"

    assert (
        model.model.support is not None and len(model.model.support) == num_classes
    ), "Support set is correct size"
    model.update_support(X, Y, 0.5)
    assert (
        model.model.support is not None and len(model.model.support) == num_classes
    ), "Support set is correct size"


@given(random_dataset=random_dataset())
@settings(deadline=None)
def test_train_episodes_with_temperature(random_dataset):
    dataset, _, way = random_dataset
    num_shot = 3
    num_episodes = 10
    episode_size = 512

    X, Y = dataset.tensors
    num_deep_features = 32
    embed_model = BasicEmbeddingModel(X.shape[1], num_deep_features)
    model = eq.EquineProtonet(embed_model, num_deep_features, use_temperature=True)
    train_dict = model.train_model(
        dataset,
        way=way,
        support_size=num_shot,
        num_episodes=num_episodes,
        episode_size=episode_size,
    )

    assert "calib_x" in train_dict
    assert "calib_y" in train_dict

    model.calibrate_temperature(train_dict["calib_x"], train_dict["calib_y"], 1, 0.01)

    # Test on multiple predictions
    eq_out = model.predict(X)
    assert len(eq_out.classes) == len(X)
    assert len(eq_out.ood_scores) == len(X)
    # Test on single prediction
    pred_out = model(X[0])
    assert len(pred_out) == 1, "Single prediction works"
    eq_out = model.predict(X[0])
    assert len(eq_out.classes) == 1, "Single prediction works"


@given(random_dataset=random_dataset())
@settings(deadline=None, max_examples=1)
def test_predict_fail_before_training(random_dataset):
    dataset, num_classes, X, embedding_model = use_basic_embedding_model(random_dataset)

    model = eq.EquineProtonet(embedding_model, num_classes)
    with pytest.raises(ValueError):
        model(X)
    with pytest.raises(ValueError):
        model.predict(X)


@given(random_dataset=random_dataset())
@settings(deadline=None, max_examples=1)
def test_equine_protonet_save_load(random_dataset) -> None:
    dataset, num_classes, X, embedding_model = use_basic_embedding_model(random_dataset)

    model = eq.EquineProtonet(embedding_model, num_classes, relative_mahal=False)
    model.train_model(dataset, num_episodes=2)

    _, tmp_filename = use_save_load_model_tests(
        model, X, tmp_filename="protonet_save_load.eq"
    )

    if os.path.exists(tmp_filename):
        os.remove(tmp_filename)  # Cleanup


@given(random_dataset=random_dataset())
@settings(deadline=None, max_examples=1)
def test_equine_protonet_save_load_with_temperature(random_dataset) -> None:
    dataset, num_classes, X, embedding_model = use_basic_embedding_model(random_dataset)

    model = eq.EquineProtonet(embedding_model, num_classes, use_temperature=True)
    model.train_model(dataset, num_episodes=2)

    new_model, tmp_filename = use_save_load_model_tests(
        model, X, tmp_filename="protonet_save_load_with_temperature.eq"
    )

    if os.path.exists(tmp_filename):
        os.remove(tmp_filename)  # Cleanup


@given(random_dataset=random_dataset())
@settings(deadline=None, max_examples=1)
def test_equine_protonet_save_load_with_feature_and_label_names(random_dataset) -> None:
    dataset, num_classes, X, embedding_model = use_basic_embedding_model(random_dataset)

    # without feature and label names
    model = eq.EquineProtonet(embedding_model, num_classes)
    model.train_model(dataset, num_episodes=2)

    new_model, tmp_filename = use_save_load_model_tests(
        model, X, tmp_filename="protonet_save_load_no_feature_and_label_names.eq"
    )

    assert new_model.get_feature_names() is None, "feature_names changed on reload"
    assert new_model.get_label_names() is None, "label_names changed on reload"

    if os.path.exists(tmp_filename):
        os.remove(tmp_filename)  # Cleanup

    # with feature and label names
    feature_names = generate_random_string_list(X.shape[1])
    label_names = generate_random_string_list(num_classes)

    model = eq.EquineProtonet(
        embedding_model,
        num_classes,
        feature_names=feature_names,
        label_names=label_names,
    )
    model.train_model(dataset, num_episodes=2)

    new_model, tmp_filename = use_save_load_model_tests(
        model, X, tmp_filename="protonet_save_load_with_feature_and_label_names.eq"
    )

    assert (
        new_model.get_feature_names() == feature_names
    ), "feature_names changed on reload"
    assert new_model.get_label_names() == label_names, "label_names changed on reload"

    if os.path.exists(tmp_filename):
        os.remove(tmp_filename)  # Cleanup



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from typing import Any, Union

import icontract
import torch
from beartype import beartype
from collections import OrderedDict
from torchmetrics.classification import (
    MulticlassAccuracy,
    MulticlassCalibrationError,
    MulticlassConfusionMatrix,
    MulticlassF1Score,
)

from .equine import Equine
from .equine_output import EquineOutput


@icontract.require(lambda y_hat, y_test: y_hat.size(dim=0) == y_test.size(dim=0))
@icontract.ensure(lambda result: result >= 0.0)
@beartype
def brier_score(y_hat: torch.Tensor, y_test: torch.Tensor) -> float:
    """
    Compute the Brier score for a multiclass problem:
    $$ \\frac{1}{N} \\sum_{i=1}^{N} \\sum_{j=1}^{M} (f_{ij} - o_{ij})^2 , $$
    where $f_{ij}$ is the predicted probability of class $j$ for inference sample $i$
    and $o_{ij}$ is the one-hot encoded ground truth label.

    Parameters
    ----------
    y_hat : torch.Tensor
        Probabilities for each class.
    y_test : torch.Tensor
        Integer argument class labels (ground truth).

    Returns
    -------
    float
        Brier score.
    """
    (_, num_classes) = y_hat.size()
    one_hot_y_test = torch.nn.functional.one_hot(y_test.long(), num_classes=num_classes)
    bs = torch.mean(torch.sum((y_hat - one_hot_y_test) ** 2, dim=1)).item()
    return bs


@icontract.require(lambda y_hat, y_test: y_hat.size(dim=0) == y_test.size(dim=0))
@icontract.ensure(lambda result: result <= 1.0)
@beartype
def brier_skill_score(y_hat: torch.Tensor, y_test: torch.Tensor) -> float:
    """
    Compute the Brier skill score as compared to randomly guessing.

    Parameters
    ----------
    y_hat : torch.Tensor
        Probabilities for each class.
    y_test : torch.Tensor
        Integer argument class labels (ground truth).

    Returns
    -------
    float
        Brier skill score.
    """
    (_, num_classes) = y_hat.size()
    random_guess = (1.0 / num_classes) * torch.ones(y_hat.size())
    bs0 = brier_score(random_guess, y_test)
    bs1 = brier_score(y_hat, y_test)
    bss = 1.0 - bs1 / bs0
    return bss


@icontract.require(lambda y_hat, y_test: y_hat.size(dim=0) == y_test.size(dim=0))
@icontract.ensure(lambda result: (0.0 <= result) and (result <= 1.0))
@beartype
def expected_calibration_error(y_hat: torch.Tensor, y_test: torch.Tensor) -> float:
    """
    Compute the expected calibration error (ECE) for a multiclass problem.

    Parameters
    ----------
    y_hat : torch.Tensor
        Probabilities for each class.
    y_test : torch.Tensor
        Class label indices (ground truth).

    Returns
    -------
    float
        Expected calibration error.
    """
    (_, num_classes) = y_hat.size()
    metric = MulticlassCalibrationError(num_classes=num_classes, n_bins=25, norm="l1")
    ece = metric(y_hat, y_test).item()
    return ece


@icontract.require(
    lambda train_y, selected_labels: len(selected_labels) <= len(train_y)
)
@icontract.ensure(
    lambda result, selected_labels: set(result.keys()).issubset(set(selected_labels))
)
@beartype
def _get_shuffle_idxs_by_class(
    train_y: torch.Tensor, selected_labels: list
) -> dict[Any, torch.Tensor]:
    """
    Internal helper function to randomly select indices of example classes for a given
    set of labels.

    Parameters
    ----------
    train_y : torch.Tensor
        Label data.
    selected_labels : list
        list of unique labels found in the label data.

    Returns
    -------
    dict[Any, torch.Tensor]
        Tensor of indices corresponding to each label.
    """
    shuffled_idxs_by_class = OrderedDict()
    for label in selected_labels:
        label_idxs = torch.argwhere(train_y == label).squeeze()
        shuffled_idxs_by_class[label] = label_idxs[torch.randperm(label_idxs.shape[0])]

    return shuffled_idxs_by_class


@icontract.require(lambda train_x, train_y: len(train_x) <= len(train_y))
@icontract.require(
    lambda selected_labels, train_x: (0 < len(selected_labels))
    & (len(selected_labels) < len(train_x))
)
@icontract.require(
    lambda support_size, train_x: (0 < support_size) & (support_size < len(train_x))
)
@icontract.require(
    lambda support_size, selected_labels, train_x: support_size * len(selected_labels)
    <= len(train_x)
)
@icontract.require(
    lambda selected_labels, shuffled_indexes: (
        (len(shuffled_indexes.keys()) == len(selected_labels))
        if shuffled_indexes is not None
        else True
    )
)
@icontract.ensure(
    lambda result, selected_labels: len(result.keys()) == len(selected_labels)
)
@beartype
def generate_support(
    train_x: torch.Tensor,
    train_y: torch.Tensor,
    support_size: int,
    selected_labels: list[Any],
    shuffled_indexes: Union[None, dict[Any, torch.Tensor]] = None,
) -> OrderedDict[int, torch.Tensor]:
    """
    Randomly select `support_size` examples of `way` classes from the examples in
    `train_x` with corresponding labels in `train_y` and return them as a dictionary.

    Parameters
    ----------
    train_x : torch.Tensor
        Input training data.
    train_y : torch.Tensor
        Corresponding classification labels.
    support_size : int
        Number of support examples for each class.
    selected_labels : list
        Selected class labels to generate examples from.
    shuffled_indexes: Union[None, dict[Any, torch.Tensor]], optional
        Simply use the precomputed indexes if they are available

    Returns
    -------
    OrderedDict[int, torch.Tensor]
        Ordered dictionary of class labels with corresponding support examples.
    """
    labels, counts = torch.unique(train_y, return_counts=True)
    if shuffled_indexes is None:
        for label, count in list(zip(labels, counts)):
            if (label in selected_labels) and (count < support_size):
                raise ValueError(f"Not enough support examples in class {label}")
        shuffled_idxs = _get_shuffle_idxs_by_class(train_y, selected_labels)
    else:
        shuffled_idxs = shuffled_indexes

    support = OrderedDict[int, torch.Tensor]()
    for label in selected_labels:
        shuffled_x = train_x[shuffled_idxs[label]]

        assert torch.unique(train_y[shuffled_idxs[label]]).tolist() == [
            label
        ], "Not enough support for label " + str(label)
        selected_support = shuffled_x[:support_size]
        support[int(label)] = selected_support

    return support


@icontract.require(lambda train_x: len(train_x.shape) >= 2)
@icontract.require(lambda train_y: len(train_y.shape) == 1)
@icontract.require(lambda support_size: support_size > 1)
@icontract.require(lambda way: way > 0)
@icontract.require(lambda episode_size: episode_size > 0)
@icontract.ensure(lambda result: len(result) == 3)
@icontract.ensure(lambda result: result[1].shape[0] == result[2].shape[0])
@icontract.ensure(lambda way, result: len(result[0]) == way)
@icontract.ensure(
    lambda support_size, result: all(
        len(support) == support_size for support in result[0].values()
    )
)
@beartype
def generate_episode(
    train_x: torch.Tensor,
    train_y: torch.Tensor,
    support_size: int,
    way: int,
    episode_size: int,
) -> tuple[OrderedDict[int, torch.Tensor], torch.Tensor, torch.Tensor]:
    """
    Generate a single episode of data for a few-shot learning task.

    Parameters
    ----------
    train_x : torch.Tensor
        Input training data.
    train_y : torch.Tensor
        Corresponding classification labels.
    support_size : int
        Number of support examples for each class.
    way : int
        Number of classes in the episode.
    episode_size : int
        Total number of examples in the episode.

    Returns
    -------
    tuple[dict[Any, torch.Tensor], torch.Tensor, torch.Tensor]
        tuple of support examples, query examples, and query labels.
    """
    labels, counts = torch.unique(train_y, return_counts=True)
    if way > len(labels):
        raise ValueError(
            f"The way (#classes in each episode), {way}, must be <= number of labels, {len(labels)}"
        )

    selected_labels = sorted(
        labels[torch.randperm(labels.shape[0])][:way].tolist()
    )  # need to be in same order every time

    for label, count in list(zip(labels, counts)):
        if (label in selected_labels) and (count < support_size):
            raise ValueError(f"Not enough support examples in class {label}")
    shuffled_idxs = _get_shuffle_idxs_by_class(train_y, selected_labels)

    support = generate_support(
        train_x, train_y, support_size, selected_labels, shuffled_idxs
    )

    examples_per_task = episode_size // way

    episode_data_list = []
    episode_label_list = []
    episode_support = OrderedDict()
    for episode_label, label in enumerate(selected_labels):
        shuffled_x = train_x[shuffled_idxs[label]]
        shuffled_y = torch.Tensor(
            [episode_label] * len(shuffled_idxs[label])
        )  # need sequential labels for episode

        num_remaining_examples = shuffled_x.shape[0] - support_size
        assert num_remaining_examples > 0, (
            "Cannot have "
            + str(num_remaining_examples)
            + " left with support_size "
            + str(support_size)
            + " and shape "
            + str(shuffled_x.shape)
            + " from train_x shaped "
            + str(train_x.shape)
        )
        episode_end_idx = support_size + min(num_remaining_examples, examples_per_task)

        episode_data_list.append(shuffled_x[support_size:episode_end_idx])
        episode_label_list.append(shuffled_y[support_size:episode_end_idx])
        episode_support[episode_label] = support[label]

    episode_x = torch.concat(episode_data_list)
    episode_y = torch.concat(episode_label_list)

    return episode_support, episode_x, episode_y.squeeze().to(torch.long)


@icontract.require(
    lambda eq_preds, true_y: eq_preds.classes.size(dim=0) == true_y.size(dim=0)
)
@beartype
def generate_model_metrics(
    eq_preds: EquineOutput, true_y: torch.Tensor
) -> dict[str, Any]:
    """
    Generate various metrics for evaluating a model's performance.

    Parameters
    ----------
    eq_preds : EquineOutput
        Model predictions.
    true_y : torch.Tensor
        True class labels.

    Returns
    -------
    dict[str, Any]
        Dictionary of model metrics.
    """
    pred_y = torch.argmax(eq_preds.classes, dim=1)
    accuracy = MulticlassAccuracy(num_classes=eq_preds.classes.shape[1])
    f1_score = MulticlassF1Score(num_classes=eq_preds.classes.shape[1], average="micro")
    confusion_matrix = MulticlassConfusionMatrix(num_classes=eq_preds.classes.shape[1])
    metrics = {
        "accuracy": accuracy(true_y, pred_y),
        "microF1Score": f1_score(true_y, pred_y),
        "confusionMatrix": confusion_matrix(true_y, pred_y).tolist(),
        "brierScore": brier_score(eq_preds.classes, true_y),
        "brierSkillScore": brier_skill_score(eq_preds.classes, true_y),
        "expectedCalibrationError": expected_calibration_error(
            eq_preds.classes, true_y
        ),
    }
    return metrics


@icontract.require(lambda Y: len(Y.shape) == 1)
@icontract.ensure(
    lambda result: all("label" in d and "numExamples" in d for d in result)
)
@icontract.ensure(lambda result: all(d["numExamples"] >= 0 for d in result))
@beartype
def get_num_examples_per_label(Y: torch.Tensor) -> list[dict[str, Any]]:
    """
    Get the number of examples per label in the given tensor.

    Parameters
    ----------
    Y : torch.Tensor
        Tensor of class labels.

    Returns
    -------
    list[dict[str, Any]]
        list of dictionaries containing label and number of examples.
    """
    tensor_labels, tensor_counts = Y.unique(return_counts=True)

    examples_per_label = []
    for i, label in enumerate(tensor_labels):
        examples_per_label.append(
            {"label": label.item(), "numExamples": tensor_counts[i].item()}
        )

    return examples_per_label


@icontract.require(lambda train_y: train_y.shape[0] > 0)
@beartype
def generate_train_summary(
    model: Equine, train_y: torch.Tensor, date_trained: str
) -> dict[str, Any]:
    """
    Generate a summary of the training data.

    Parameters
    ----------
    model : Equine
        Model object.
    train_y : torch.Tensor
        Training labels.
    date_trained : str
        Date of training.

    Returns
    -------
    dict[str, Any]
        Dictionary containing training summary.
    """
    train_summary = {
        "numTrainExamples": get_num_examples_per_label(train_y),
        "dateTrained": date_trained,
        "modelType": model.__class__.__name__,
    }
    return train_summary


@icontract.require(
    lambda eq_preds, test_y: test_y.shape[0] == eq_preds.classes.shape[0]
)
@beartype
def generate_model_summary(
    model: Equine,
    eq_preds: EquineOutput,
    test_y: torch.Tensor,
) -> dict[str, Any]:
    """
    Generate a summary of the model's performance.

    Parameters
    ----------
    model : Equine
        Model object.
    eq_preds : EquineOutput
        Model predictions.
    test_y : torch.Tensor
        True class labels.

    Returns
    -------
    dict[str, Any]
        Dictionary containing model summary.
    """
    summary = generate_model_metrics(eq_preds, test_y)
    summary["numTestExamples"] = get_num_examples_per_label(test_y)
    summary.update(model.train_summary)  # union of train_summary and generated metrics

    return summary


@icontract.require(lambda cov: cov.shape[-2] == cov.shape[-1])
def mahalanobis_distance_nosq(x: torch.Tensor, cov: torch.Tensor) -> torch.Tensor:
    """
    Compute Mahalanobis distance $x^T C x$ (without square root), assume cov is symmetric positive definite

    Parameters
    ----------
    x : torch.Tensor
        vectors to compute distances for
    cov : torch.Tensor
        covariance matrix, assumes first dimension is number of classes
    """
    U, S, _ = torch.linalg.svd(cov)
    S_inv_sqrt = torch.stack(
        [torch.diag(torch.sqrt(1.0 / S[i])) for i in range(S.shape[0])], dim=0
    )
    prod = torch.matmul(S_inv_sqrt, torch.transpose(U, 1, 2))
    dist = torch.sum(torch.square(torch.matmul(prod, x)), dim=1)
    return dist


@icontract.require(
    lambda X, Y: X.shape[0] == Y.shape[0],
    "X and Y must have the same number of samples.",
)
@icontract.require(
    lambda test_size: 0.0 < test_size < 1.0, "test_size must be between 0 and 1."
)
@icontract.ensure(
    lambda result: len(result) == 4, "Function must return four elements."
)
@icontract.ensure(
    lambda X, result: result[0].shape[0] + result[1].shape[0] == X.shape[0],
    "Total samples must be preserved.",
)
@icontract.ensure(
    lambda Y, result: result[2].shape[0] + result[3].shape[0] == Y.shape[0],
    "Total labels must be preserved.",
)
@icontract.ensure(
    lambda result: result[0].shape[0] == result[2].shape[0],
    "Train features and labels must match in size.",
)
@icontract.ensure(
    lambda result: result[1].shape[0] == result[3].shape[0],
    "Test features and labels must match in size.",
)
@beartype
def stratified_train_test_split(
    X: torch.Tensor, Y: torch.Tensor, test_size: float
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    A pytorch-ified version of sklearn's train_test_split with data stratification

    Parameters
    ----------
    X : torch.Tensor
        Input features tensor of shape (n_samples, n_features).
    Y : torch.Tensor
        Labels tensor of shape (n_samples,).
    test_size : float
        Proportion of the dataset to include in the test split (between 0.0 and 1.0).

    Returns
    -------
    train_x : torch.Tensor
        Training set features.
    calib_x : torch.Tensor
        Test set features.
    train_y : torch.Tensor
        Training set labels.
    calib_y : torch.Tensor
        Test set labels.
    """
    unique_classes, class_counts = torch.unique(Y, return_counts=True)
    test_counts = (class_counts.float() * test_size).round().long()
    train_indices = []
    test_indices = []

    for cls, test_count in zip(unique_classes, test_counts):
        cls_indices = torch.where(Y == cls)[0]
        cls_indices = cls_indices[torch.randperm(len(cls_indices))]
        test_idx = cls_indices[:test_count]
        train_idx = cls_indices[test_count:]
        train_indices.append(train_idx)
        test_indices.append(test_idx)

    train_indices = torch.cat(train_indices)
    test_indices = torch.cat(test_indices)

    train_x = X[train_indices]
    train_y = Y[train_indices]
    calib_x = X[test_indices]
    calib_y = Y[test_indices]

    return train_x, calib_x, train_y, calib_y



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

import torch

from .equine import Equine
from .equine_gp import EquineGP
from .equine_protonet import EquineProtonet


def load_equine_model(model_path: str) -> Equine:
    """
    Attempt to load an EQUINE model from a file

    Parameters
    ----------
    model_path : str
        The path to the model file

    Returns
    -------
    Equine
        The loaded EQUINE model

    Raises
    ------
    ValueError
        If the model type is unknown
    """
    model_type = torch.load(model_path, weights_only=False)["train_summary"][
        "modelType"
    ]

    if model_type == "EquineProtonet":
        model = EquineProtonet.load(model_path)
    elif model_type == "EquineGP":
        model = EquineGP.load(model_path)
    else:
        raise ValueError(f"Unknown model type '{model_type}'")
    return model



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from torch import Tensor


@dataclass
class EquineOutput:
    """
    Output object containing prediction probabilities, OOD scores, and embeddings, which can be used for visualization.

    Attributes
    ----------
    classes : Tensor
        Tensor of predicted class probabilities.
    ood_scores : Tensor
        Tensor of out-of-distribution (OOD) scores.
    embeddings : Tensor
        Tensor of embeddings produced by the model.
    """

    classes: Tensor
    ood_scores: Tensor
    embeddings: Tensor
